"""
Contributors: Alexander Jüstel, Arthur Endlein Correia, Florian Wellmann

GemGIS is a Python-based, open-source geographic information processing library.
It is capable of preprocessing spatial data such as vector data (shape files, geojson files,
geopackages), raster data (tif, png,...), data obtained from web services (WMS, WFS, WCS) or XML/KML
files. Preprocessed data can be stored in a dedicated Data Class to be passed to the geomodeling package
GemPy in order to accelerate to model building process. In addition, enhanced 3D visualization of data is
powered by the PyVista package.

GemGIS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

GemGIS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (LICENSE.md) for more details.

"""

import pytest
import rasterio
import numpy as np
import geopandas as gpd
import shapely
import pandas as pd
import affine

# Definition of GeoDataFrames
###########################################################
points1 = shapely.geometry.point.Point(19.150128045807676, 293.313485355882)

points2 = shapely.geometry.point.Point(61.93436666575576, 381.4593263680641)

points3 = shapely.geometry.point.Point(109.3578600758187, 480.9455679783049)

points4 = shapely.geometry.point.Point(157.812298994796, 615.9994296460927)

points5 = shapely.geometry.point.Point(191.3180280345144, 719.0939805375339)

gdf_interfaces1_points = gpd.GeoDataFrame(geometry=[points1, points2, points3, points4, points5], crs='EPSG:4326')
gdf_interfaces1_points['formation'] = 'Ton'
gdf_interfaces1_points['id'] = None

points1 = shapely.geometry.point.Point(0, 0)
points2 = shapely.geometry.point.Point(0, 500)
points3 = shapely.geometry.point.Point(500, 0)
points4 = shapely.geometry.point.Point(500, 500)

gdf_clipping_points = gpd.GeoDataFrame(geometry=[points1, points2, points3, points4], crs='EPSG:4326')
gdf_clipping_points['id'] = None
gdf_clipping_points['X'] = [0, 0, 500, 500]
gdf_clipping_points['Y'] = [0, 500, 0, 500]


# Testing sample_from_array
###########################################################
@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_from_array(dem):
    from gemgis.raster import sample_from_array

    extent = [0, 972.0, 0, 1069.0]

    point_x = [800, 400]
    point_y = [100, 1000]

    samples = sample_from_array(dem.read(1), extent, point_x, point_y)

    assert isinstance(samples, np.ndarray)
    assert len(samples) == 2
    assert samples[0] == 276.4043273925781
    assert samples[1] == 739.3092041015625


@pytest.mark.parametrize("gdf_interfaces1_points", [gdf_interfaces1_points])
@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_from_array(gdf_interfaces1_points, dem):
    from gemgis.raster import sample_from_array
    from gemgis.vector import extract_xy

    extent = [0, 972.0, 0, 1069.0]

    gdf = extract_xy(gdf_interfaces1_points)
    point_x = gdf['X'].tolist()[:5]
    point_y = gdf['Y'].tolist()[:5]

    samples = sample_from_array(dem.read(1), extent, point_x, point_y)

    assert isinstance(samples, np.ndarray)
    assert len(samples) == 5
    assert isinstance(samples[0], np.float32)
    assert samples[0] == np.float32(366.61255)
    assert samples[1] == np.float32(402.09912)
    assert samples[2] == np.float32(460.6181)
    assert samples[3] == np.float32(529.0156)
    assert samples[4] == np.float32(597.6325)


@pytest.mark.parametrize("array",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_sample(array):
    from gemgis.raster import sample_from_array
    sample = sample_from_array(array, extent=[1000, 2069, 1000, 1972], point_x=[1500], point_y=[1500])

    assert array.ndim == 2
    assert array.shape == (1069, 972)
    assert isinstance(sample, float)
    assert sample == 573.9062885108234


@pytest.mark.parametrize("array",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_sample_error(array):
    from gemgis.raster import sample_from_array
    with pytest.raises(TypeError):
        sample_from_array(list(array), [1000, 2069, 1000, 1972], point_x=1500, point_y=1500)
    with pytest.raises(TypeError):
        sample_from_array(list(array), (1000, 2069, 1000, 1972), point_x=1500, point_y=1500)
    with pytest.raises(ValueError):
        sample_from_array(array, [1000, 2069, 1000], point_x=1500, point_y=1500)
    with pytest.raises(ValueError):
        sample_from_array(array, [1000, 2069, 1000, 1972, 500], point_x=1500, point_y=1500)
    with pytest.raises(TypeError):
        sample_from_array(array, [1000, 2069, 1000, 1972], point_x=(1500, 1500), point_y=5)
    with pytest.raises(ValueError):
        sample_from_array(array, [1000, 2069, 1000, 1972], [1500, 1500, 1500], [1500])
    with pytest.raises(ValueError):
        sample_from_array(array, [1000, 2069, 1000, '1972'], [1500, 1500], [1500])
    with pytest.raises(ValueError):
        sample_from_array(array, [1000, 2069, 1000, 1972], [1500, '1500'], [1500])
    with pytest.raises(ValueError):
        sample_from_array(array, [1000, 2069, 1000, 1972], [15000, 1500], [1500])
    with pytest.raises(ValueError):
        sample_from_array(array, [1000, 2069, 1000, 1972], [1500, 15000], [1500])
    with pytest.raises(ValueError):
        sample_from_array(array, [1000, 2069, 1000, 1972], [150, 1500], [1500])
    with pytest.raises(ValueError):
        sample_from_array(array, [1000, 2069, 1000, 1972], [1500, 150], [1500])


# Testing sample_from_rasterio
###########################################################
@pytest.mark.parametrize("gdf_interfaces1_points", [gdf_interfaces1_points])
@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_from_rasterio(gdf_interfaces1_points, dem):
    from gemgis.raster import sample_from_rasterio
    from gemgis.vector import extract_xy

    gdf = extract_xy(gdf_interfaces1_points)
    point_x = gdf['X'].tolist()[:5]
    point_y = gdf['Y'].tolist()[:5]

    samples = sample_from_rasterio(dem, point_x, point_y)

    assert isinstance(samples, list)
    assert len(samples) == 5
    assert isinstance(samples[0], float)
    assert samples[0] == 364.994873046875
    assert samples[1] == 400.3435974121094
    assert samples[2] == 459.54931640625
    assert samples[3] == 525.6910400390625
    assert samples[4] == 597.6325073242188


@pytest.mark.parametrize("gdf_interfaces1_points", [gdf_interfaces1_points])
@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_from_rasterio2(gdf_interfaces1_points, dem):
    from gemgis.raster import sample_from_rasterio
    from gemgis.vector import extract_xy

    gdf = extract_xy(gdf_interfaces1_points)
    point_x = gdf['X'].tolist()[0]
    point_y = gdf['Y'].tolist()[0]

    samples = sample_from_rasterio(dem, point_x, point_y)

    assert isinstance(samples, float)
    assert samples == 364.994873046875


# Testing sample_randomly
###########################################################

@pytest.mark.parametrize("array",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_sample_randomly_go(array):
    from gemgis.raster import sample_randomly
    random_sample = sample_randomly(array, extent=[1000, 2069, 1000, 1972], seed=1)

    assert array.ndim == 2
    assert array.shape == (1069, 972)
    assert isinstance(random_sample[0], float)
    assert isinstance(random_sample[1], list)
    assert random_sample == (518.1631561814993, [1445.7965230270515, 1700.1554076257776])
    assert all(isinstance(n, float) for n in random_sample[1])


@pytest.mark.parametrize("array",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_sample_randomly_error(array):
    from gemgis.raster import sample_randomly
    with pytest.raises(TypeError):
        sample_randomly([array], [1000, 2069, 1000, 1972], seed=1)
    with pytest.raises(TypeError):
        sample_randomly(array, (1000, 2069, 1000, 1972), seed=1)
    with pytest.raises(TypeError):
        sample_randomly(array, [1000, 2069, 1000, 1972], seed=1.0)
    with pytest.raises(TypeError):
        sample_randomly(array, [1000, 2069, 1000, '1972'], seed=1)


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_randomly_array(dem):
    from gemgis.raster import sample_randomly

    sample = sample_randomly(dem.read(1), extent=[1000, 2069, 1000, 1972], seed=1)

    assert sample == (668.4022827148438, [1445.7965230270515, 1700.1554076257776])
    assert sample[0] == 668.4022827148438
    assert isinstance(sample[0], float)
    assert isinstance(sample[1], list)
    assert isinstance(sample[1][0], float)
    assert isinstance(sample[1][1], float)


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_randomly_array_multiple(dem):
    from gemgis.raster import sample_randomly

    sample, coordinates = sample_randomly(dem.read(1), n=2, extent=[1000, 2069, 1000, 1972], seed=1)

    assert isinstance(sample, list)
    assert isinstance(coordinates, list)

    assert isinstance(sample[0], float)
    assert isinstance(sample[1], float)
    assert isinstance(coordinates[0][0], float)
    assert isinstance(coordinates[0][1], float)

    assert sample[0] == 416.4200439453125
    assert sample[1] == 402.33673095703125
    assert coordinates[0][0] == 1445.7965230270515
    assert coordinates[0][1] == 1770.026883489667
    assert coordinates[1][0] == 1000.1111723224592
    assert coordinates[1][1] == 1293.8672605981483


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_randomly_raster(dem):
    from gemgis.raster import sample_randomly

    sample = sample_randomly(dem, extent=[1000, 2069, 1000, 1972], seed=1)

    assert sample == (668.4022827148438, [404.92957493148504, 769.3808873834653])
    assert sample[0] == 668.4022827148438
    assert isinstance(sample[0], float)
    assert isinstance(sample[1], list)
    assert isinstance(sample[1][0], float)
    assert isinstance(sample[1][1], float)


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_randomly_raster_multiple(dem):
    from gemgis.raster import sample_randomly

    sample, coordinates = sample_randomly(dem, n=2, extent=[1000, 2069, 1000, 1972], seed=1)

    assert isinstance(sample, list)
    assert isinstance(coordinates, list)

    assert isinstance(sample[0], float)
    assert isinstance(sample[1], float)
    assert isinstance(coordinates[0][0], float)
    assert isinstance(coordinates[0][1], float)

    assert sample[0] == 416.4200439453125
    assert sample[1] == 402.33673095703125
    assert coordinates[0][0] == 404.92957493148504
    assert coordinates[0][1] == 699.4371703486034
    assert coordinates[1][0] == 0.12216410696185687
    assert coordinates[1][1] == 322.92238447267215


# Testing calculate_hillshade
###########################################################
@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_calculate_hillshades_array(dem):
    from gemgis.raster import calculate_hillshades

    hillshades = calculate_hillshades(dem)

    assert dem.read(1).ndim == 2
    assert dem.read(1).shape == (275, 250)
    assert isinstance(dem, rasterio.io.DatasetReader)
    assert isinstance(dem.read(1), np.ndarray)
    assert dem.read(1).ndim == 2
    assert isinstance(hillshades, np.ndarray)
    assert hillshades.ndim == 2


@pytest.mark.parametrize("dem",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_calculate_hillshades_array2(dem):
    from gemgis.raster import calculate_hillshades

    hillshades = calculate_hillshades(dem, [0, 972, 0, 1069])

    assert dem.ndim == 2
    assert dem.shape == (1069, 972)
    assert isinstance(dem, np.ndarray)
    assert dem.ndim == 2
    assert isinstance(hillshades, np.ndarray)
    assert hillshades.ndim == 2


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_calculate_hillshades_raster(dem):
    from gemgis.raster import calculate_hillshades

    hillshades = calculate_hillshades(dem)

    assert dem.read(1).ndim == 2
    assert dem.read(1).shape == (275, 250)
    assert isinstance(dem, rasterio.io.DatasetReader)
    assert isinstance(dem.read(1), np.ndarray)
    assert dem.read(1).ndim == 2
    assert isinstance(hillshades, np.ndarray)
    assert hillshades.ndim == 2


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_calculate_hillshades_error(dem):
    from gemgis.raster import calculate_hillshades

    with pytest.raises(TypeError):
        calculate_hillshades([dem])

    with pytest.raises(TypeError):
        calculate_hillshades(dem, altdeg=[5], azdeg=10)

    with pytest.raises(TypeError):
        calculate_hillshades(dem, altdeg=5, azdeg=[10])

    with pytest.raises(ValueError):
        calculate_hillshades(dem, altdeg=-5, azdeg=10)

    with pytest.raises(ValueError):
        calculate_hillshades(dem, altdeg=100, azdeg=10)

    with pytest.raises(ValueError):
        calculate_hillshades(dem, altdeg=45, azdeg=-5)

    with pytest.raises(ValueError):
        calculate_hillshades(dem, altdeg=45, azdeg=400)


# Testing calculate_slope
###########################################################
@pytest.mark.parametrize("raster",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_calculate_slope(raster):
    from gemgis.raster import calculate_slope

    slope = calculate_slope(raster)

    assert isinstance(raster, rasterio.io.DatasetReader)
    assert raster.read(1).shape == (1000, 1000)
    assert raster.read(1).ndim == 2
    assert isinstance(slope, np.ndarray)
    assert slope.ndim == 2
    assert slope[0][0] == 45
    for i in np.arange(0, 1000, 100):
        for j in np.arange(0, 1000, 100):
            assert round(slope[i][j], 10) == 45
    assert slope.shape == (1000, 1000)


@pytest.mark.parametrize("array",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_calculate_slope_array(array):
    from gemgis.raster import calculate_slope

    slope = calculate_slope(array, [0, 972, 0, 1069])

    assert isinstance(array, np.ndarray)
    assert array.shape == (1069, 972)
    assert array.ndim == 2
    assert isinstance(slope, np.ndarray)
    assert slope.ndim == 2
    assert slope[0][0] == 11.598369665181522
    assert slope.shape == (1069, 972)


@pytest.mark.parametrize("raster",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_calculate_slope_raster(raster):
    from gemgis.raster import calculate_slope

    slope = calculate_slope(raster)

    assert isinstance(raster, rasterio.io.DatasetReader)
    assert raster.read(1).shape == (275, 250)
    assert raster.read(1).ndim == 2
    assert isinstance(slope, np.ndarray)
    assert slope.ndim == 2
    assert slope.shape == (275, 250)


@pytest.mark.parametrize("raster",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_calculate_slope(raster):
    from gemgis.raster import calculate_slope

    with pytest.raises(TypeError):
        calculate_slope([raster])


# Testing calculate_aspect
###########################################################
@pytest.mark.parametrize("raster",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_calculate_aspect(raster):
    from gemgis.raster import calculate_aspect

    aspect = calculate_aspect(raster)

    assert isinstance(raster, rasterio.io.DatasetReader)
    assert raster.read(1).shape == (275, 250)
    assert raster.read(1).ndim == 2
    assert isinstance(aspect, np.ndarray)
    assert aspect.ndim == 2
    assert round(aspect[0][0]) == 246
    assert aspect.shape == (275, 250)


@pytest.mark.parametrize("raster",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_calculate_aspect_array(raster):
    from gemgis.raster import calculate_aspect

    aspect = calculate_aspect(raster, [0, 972, 0, 1069])

    assert isinstance(raster, np.ndarray)
    assert raster.shape == (1069, 972)
    assert raster.ndim == 2
    assert isinstance(aspect, np.ndarray)
    assert aspect.ndim == 2
    assert aspect[0][0] == 140.833617575023
    assert aspect.shape == (1069, 972)


@pytest.mark.parametrize("raster",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_calculate_aspect_error(raster):
    from gemgis.raster import calculate_aspect

    with pytest.raises(TypeError):
        calculate_aspect([raster])


# Testing clip_raster_data_by_shape
###########################################################
@pytest.mark.parametrize("raster",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_clip_by_shape(raster):
    from gemgis.raster import clip_by_polygon

    polygon = shapely.geometry.Polygon([(0, 0), (0, 500), (500, 500), (500, 0)])

    clipped_array = clip_by_polygon(raster=raster,
                                    polygon=polygon)

    assert raster.read(1).ndim == 2
    assert raster.read(1).shape == (275, 250)
    assert isinstance(raster, rasterio.io.DatasetReader)
    assert isinstance(clipped_array, np.ndarray)
    assert clipped_array.ndim == 2
    assert clipped_array.shape == (129, 129)


@pytest.mark.parametrize("raster",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_clip_by_shape_array(raster):
    from gemgis.raster import clip_by_polygon

    polygon = shapely.geometry.Polygon([(0, 0), (0, 500), (500, 500), (500, 0)])

    clipped_array = clip_by_polygon(raster=raster,
                                    polygon=polygon,
                                    raster_extent=[0, 972, 0, 1069])

    assert raster.ndim == 2
    assert raster.shape == (1069, 972)
    assert isinstance(raster, np.ndarray)
    assert isinstance(clipped_array, np.ndarray)
    assert clipped_array.ndim == 2
    assert clipped_array.shape == (500, 500)


@pytest.mark.parametrize("raster",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
@pytest.mark.parametrize("gdf_clipping_points", [gdf_clipping_points])
def test_clip_by_shape_error(raster, gdf_clipping_points):
    from gemgis.raster import clip_by_polygon

    with pytest.raises(TypeError):
        clip_by_polygon([raster], gdf_clipping_points)
    with pytest.raises(TypeError):
        clip_by_polygon(raster, [gdf_clipping_points])
    with pytest.raises(TypeError):
        clip_by_polygon(raster, gdf_clipping_points, save_clipped_raster='True')


# Testing sample_orientations_from_raster
###########################################################
@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_orientations_from_raster(dem):
    from gemgis.raster import sample_orientations

    orientations = sample_orientations(raster=dem,
                                       random_samples=5,
                                       formation='surface')

    assert isinstance(orientations, gpd.geodataframe.GeoDataFrame)
    assert {'X', 'Y', 'Z', 'formation', 'dip', 'azimuth', 'polarity'}.issubset(orientations.columns)
    assert len(orientations) == 5


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_orientations_from_raster_point(dem):
    from gemgis.raster import sample_orientations
    from gemgis.utils import set_extent

    extent = set_extent(0, 972, 0, 1069)

    orientations = sample_orientations(raster=dem.read(1),
                                       extent=extent,
                                       point_x=500,
                                       point_y=500,
                                       formation='surface')

    assert isinstance(orientations, gpd.geodataframe.GeoDataFrame)
    assert {'X', 'Y', 'Z', 'formation', 'dip', 'azimuth', 'polarity'}.issubset(orientations.columns)
    assert len(orientations) == 1


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_orientations_from_raster_points(dem):
    from gemgis.raster import sample_orientations
    from gemgis.utils import set_extent

    extent = set_extent(0, 972, 0, 1069)

    orientations = sample_orientations(raster=dem.read(1),
                                       extent=extent,
                                       point_x=[500, 500],
                                       point_y=[600, 600],
                                       formation='surface')

    assert isinstance(orientations, pd.DataFrame)
    assert {'X', 'Y', 'Z', 'formation', 'dip', 'azimuth', 'polarity'}.issubset(orientations.columns)
    assert len(orientations) == 2


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_orientations_from_raster_points3(dem):
    from gemgis.raster import sample_orientations
    from gemgis.utils import set_extent

    extent = set_extent(0, 972, 0, 1069)

    orientations = sample_orientations(raster=dem.read(1),
                                       extent=extent,
                                       point_x=[500, 600, 700],
                                       point_y=[500, 600, 700],
                                       formation='surface')

    assert isinstance(orientations, pd.DataFrame)
    assert {'X', 'Y', 'Z', 'formation', 'dip', 'azimuth', 'polarity'}.issubset(orientations.columns)
    assert len(orientations) == 3


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_orientations_from_raster_error(dem):
    from gemgis.raster import sample_orientations
    from gemgis.utils import set_extent

    extent = set_extent(0, 972, 0, 1069)

    with pytest.raises(TypeError):
        sample_orientations(raster=[dem],
                            extent=extent,
                            point_x=[500, 600, 700],
                            point_y=[500, 600, 700],
                            formation='surface')
    with pytest.raises(TypeError):
        sample_orientations(raster=dem,
                            extent=[extent],
                            point_x=[500, 600, 700],
                            point_y=[500, 600, 700],
                            formation=['surface'])
    with pytest.raises(TypeError):
        sample_orientations(raster=dem,
                            extent=extent,
                            point_x=(500, 600, 700),
                            point_y=[500, 600, 700],
                            formation='surface')
    with pytest.raises(TypeError):
        sample_orientations(raster=dem,
                            extent=extent,
                            point_x=[500, 600, 700],
                            point_y=[500, 600, 700],
                            formation=['surface'])
    with pytest.raises(TypeError):
        sample_orientations([dem], extent, formation='surface')


# Testing sample_interfaces_from_raster
###########################################################
@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_interfaces_from_raster(dem):
    from gemgis.raster import sample_interfaces

    interfaces = sample_interfaces(raster=dem,
                                   random_samples=5,
                                   formation='surface')

    assert isinstance(interfaces, gpd.geodataframe.GeoDataFrame)
    assert {'X', 'Y', 'Z', 'formation'}.issubset(interfaces.columns)
    assert len(interfaces) == 5


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_interfaces_from_raster_point(dem):
    from gemgis.raster import sample_interfaces
    from gemgis.utils import set_extent

    extent = set_extent(0, 972, 0, 1069)

    interfaces = sample_interfaces(raster=dem.read(1),
                                   extent=extent,
                                   point_x=500,
                                   point_y=500,
                                   formation='surface')

    assert isinstance(interfaces, gpd.geodataframe.GeoDataFrame)
    assert {'X', 'Y', 'Z', 'formation'}.issubset(interfaces.columns)
    assert len(interfaces) == 1


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_interfaces_from_raster_points(dem):
    from gemgis.raster import sample_interfaces
    from gemgis.utils import set_extent

    extent = set_extent(0, 972, 0, 1069)

    interfaces = sample_interfaces(raster=dem.read(1),
                                   extent=extent,
                                   point_x=[500, 500],
                                   point_y=[600, 600],
                                   formation='surface')

    assert isinstance(interfaces, pd.DataFrame)
    assert {'X', 'Y', 'Z', 'formation'}.issubset(interfaces.columns)
    assert len(interfaces) == 2


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_interfaces_from_raster_points3(dem):
    from gemgis.raster import sample_interfaces
    from gemgis.utils import set_extent

    extent = set_extent(0, 972, 0, 1069)

    interfaces = sample_interfaces(raster=dem.read(1),
                                   extent=extent,
                                   point_x=[500, 600, 700],
                                   point_y=[500, 600, 700],
                                   formation='surface')

    assert isinstance(interfaces, pd.DataFrame)
    assert {'X', 'Y', 'Z', 'formation'}.issubset(interfaces.columns)
    assert len(interfaces) == 3


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_sample_interfaces_from_raster_error(dem):
    from gemgis.raster import sample_interfaces
    from gemgis.utils import set_extent

    extent = set_extent(0, 972, 0, 1069)

    with pytest.raises(TypeError):
        sample_interfaces(raster=[dem],
                          extent=extent,
                          point_x=[[500, 500], [600, 600], [700, 700]],
                          point_y=[[500, 500], [600, 600], [700, 700]],
                          formation='surface')
    with pytest.raises(TypeError):
        sample_interfaces(raster=dem,
                          extent=[extent],
                          point_x=[[500, 500], [600, 600], [700, 700]],
                          point_y=[[500, 500], [600, 600], [700, 700]],
                          formation='surface')
    with pytest.raises(TypeError):
        sample_interfaces(raster=dem,
                          extent='extent',
                          point_x=([500, 500], [600, 600], [700, 700]),
                          point_y=[[500, 500], [600, 600], [700, 700]],
                          formation='surface')
    with pytest.raises(TypeError):
        sample_interfaces(raster=dem,
                          extent=extent,
                          point_x=[[500, 500], [600, 600], [700, 700]],
                          point_y=[[500, 500], [600, 600], [700, 700]],
                          formation=['surface'])
    with pytest.raises(TypeError):
        sample_interfaces(raster=[dem],
                          extent=extent,
                          formation='surface')


# Testing resize_raster_by_array
###########################################################
@pytest.mark.parametrize("array1",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
@pytest.mark.parametrize("array2",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_resize_by_array(array1, array2):
    from gemgis.raster import resize_by_array

    array_rescaled = resize_by_array(raster=array1,
                                     array=array2)

    assert array1.read(1).ndim == 2
    assert array1.read(1).shape == (275, 250)

    assert array2.ndim == 2
    assert array2.shape == (1069, 972)

    assert array_rescaled.ndim == 2
    assert array_rescaled.shape == (1069, 972)


@pytest.mark.parametrize("array2",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
@pytest.mark.parametrize("array1",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_resize_by_array_2(array1, array2):
    from gemgis.raster import resize_by_array

    array_rescaled = resize_by_array(raster=array1,
                                     array=array2)

    assert array2.read(1).ndim == 2
    assert array2.read(1).shape == (275, 250)

    assert array1.ndim == 2
    assert array1.shape == (1069, 972)

    assert array_rescaled.ndim == 2
    assert array_rescaled.shape == (275, 250)


@pytest.mark.parametrize("array1",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
@pytest.mark.parametrize("array2",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_resize_by_array_error(array1, array2):
    from gemgis.raster import resize_by_array

    with pytest.raises(TypeError):
        resize_by_array(raster=[array1],
                        array=array2)

    with pytest.raises(TypeError):
        resize_by_array(raster=array1.read(1),
                        array=[array2])


# Testing resize_raster
###########################################################
@pytest.mark.parametrize("array1",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_resize_raster(array1):
    from gemgis.raster import resize_raster

    array_rescaled = resize_raster(raster=array1,
                                   width=500,
                                   height=500)

    assert array1.read(1).ndim == 2
    assert array1.read(1).shape == (275, 250)

    assert array_rescaled.ndim == 2
    assert array_rescaled.shape == (500, 500)


@pytest.mark.parametrize("array1",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_resize_raster_array(array1):
    from gemgis.raster import resize_raster

    array_rescaled = resize_raster(raster=array1,
                                   width=500,
                                   height=500)

    assert array1.ndim == 2
    assert array1.shape == (1069, 972)

    assert array_rescaled.ndim == 2
    assert array_rescaled.shape == (500, 500)


@pytest.mark.parametrize("array1",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
@pytest.mark.parametrize("array2",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_rescale_raster_error(array1, array2):
    from gemgis.raster import resize_raster

    with pytest.raises(TypeError):
        resize_raster(raster=[array1.read(1)],
                      width=500,
                      height=500)

    with pytest.raises(TypeError):
        resize_raster(raster=array1.read(1),
                      width='500',
                      height='500')

    with pytest.raises(TypeError):
        resize_raster(raster=[array2],
                      width=500,
                      height=500)

    with pytest.raises(TypeError):
        resize_raster(raster=array2,
                      width=[500],
                      height=[500])


# Testing calculate_difference
###########################################################
def test_calculate_difference():
    from gemgis.raster import calculate_difference

    array_diff = calculate_difference(raster1=np.ones(9).reshape(3, 3),
                                      raster2=np.zeros(9).reshape(3, 3))
    assert array_diff.ndim == 2
    assert array_diff.shape == (3, 3)
    for i in range(array_diff.shape[1]):
        for j in range(array_diff.shape[0]):
            assert array_diff[j][i] == 1


# Testing calculate_difference
###########################################################
@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_calculate_difference(dem):
    from gemgis.raster import calculate_difference

    array_diff = calculate_difference(raster1=dem.read(1),
                                      raster2=dem.read(1) + 5,
                                      flip_array=False)
    assert array_diff.ndim == 2
    assert array_diff.shape == (275, 250)
    for i in range(array_diff.shape[1]):
        for j in range(array_diff.shape[0]):
            assert round(array_diff[j][i]) == -5


# Testing calculate_difference
###########################################################
def test_calculate_difference_error():
    from gemgis.raster import calculate_difference

    with pytest.raises(TypeError):
        calculate_difference(raster1=[np.ones(9).reshape(3, 3)],
                             raster2=np.zeros(9).reshape(3, 3))
    with pytest.raises(TypeError):
        calculate_difference(raster1=np.ones(9).reshape(3, 3),
                             raster2=[np.zeros(9).reshape(3, 3)])


# Testing read_msh
###########################################################
def test_read_msh():
    from gemgis.raster import read_msh

    data = read_msh('../../gemgis/tests/data/GM_Breccia.msh')

    assert isinstance(data, dict)
    assert 'Tri' in data
    assert 'Location' in data
    assert isinstance(data['Tri'], np.ndarray)
    assert isinstance(data['Location'], np.ndarray)


# Testing create_filepaths
###########################################################
def test_create_filepaths():
    from gemgis.raster import create_filepaths

    paths = create_filepaths('../../gemgis/tests/data/', search_criteria='raster1*.tif')

    assert isinstance(paths, list)
    assert paths == ['C:\\Users\\ale93371\\Documents\\gemgis\\tests\\data\\raster1.tif']
    assert isinstance(paths[0], str)


# Testing create_filepaths
###########################################################
def test_create_src_list():
    from gemgis.raster import create_src_list, create_filepaths

    paths = create_filepaths('../../gemgis/tests/data/', search_criteria='raster1*.tif')
    source_paths = create_src_list(dirpath='', search_criteria='', filepaths=paths)

    assert isinstance(paths, list)
    assert paths == ['C:\\Users\\ale93371\\Documents\\gemgis\\tests\\data\\raster1.tif']

    assert isinstance(source_paths, list)
    assert isinstance(source_paths[0], rasterio.io.DatasetReader)
    assert source_paths[0].name == 'C:/Users/ale93371/Documents/gemgis/tests/data/raster1.tif'


# Testing merge_tiles
###########################################################
def test_merge_tiles():
    from gemgis.raster import merge_tiles, create_src_list, create_filepaths

    paths = create_filepaths('../../gemgis/tests/data/', search_criteria='raster1*.tif')
    source_paths = create_src_list(dirpath='', search_criteria='', filepaths=paths)

    assert isinstance(paths, list)
    assert paths == ['C:\\Users\\ale93371\\Documents\\gemgis\\tests\\data\\raster1.tif']

    assert isinstance(source_paths, list)
    assert isinstance(source_paths[0], rasterio.io.DatasetReader)
    assert source_paths[0].name == 'C:/Users/ale93371/Documents/gemgis/tests/data/raster1.tif'

    mosaic, transform = merge_tiles(src_files=source_paths)

    assert isinstance(mosaic, np.ndarray)
    assert isinstance(transform, affine.Affine)


# Testing save_array_as_tiff
###########################################################
@pytest.mark.parametrize("raster",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_save_raster_as_tiff(raster):
    from gemgis.raster import save_as_tiff

    save_as_tiff('test', raster, [0, 1069, 0, 972], 'EPSG:4326')

    assert raster.ndim == 2
    assert raster.shape == (1069, 972)
    assert isinstance(raster, np.ndarray)


@pytest.mark.parametrize("raster",
                         [
                             np.load('../../gemgis/tests/data/array_rbf.npy')
                         ])
def test_save_raster_as_tiff(raster):
    from gemgis.raster import save_as_tiff

    with pytest.raises(TypeError):
        save_as_tiff(['test'], raster, [0, 1069, 0, 972], 'EPSG:4326')
    with pytest.raises(TypeError):
        save_as_tiff('test', [raster], [0, 1069, 0, 972], 'EPSG:4326')
    with pytest.raises(TypeError):
        save_as_tiff('test', raster, (0, 1069, 0, 972), 'EPSG:4326')
    with pytest.raises(TypeError):
        save_as_tiff('test', raster, [0, 1069, 0, 972], ['EPSG:4326'])


# Testing read_asc
###########################################################
def test_read_asc():
    from gemgis.raster import read_asc

    data = read_asc(path='../../gemgis/tests/data/top_dinant_final_tvd.asc')

    assert isinstance(data['Data'], np.ndarray)
    assert isinstance(data['Extent'], list)
    assert isinstance(data['Resolution'], float)
    assert isinstance(data['Nodata_val'], float)


# Testing read_asc
###########################################################
def test_read_zmap():
    from gemgis.raster import read_zmap

    data = read_zmap(path='../../gemgis/tests/data/top_dinant_final_tvd.dat')

    assert isinstance(data['Data'], np.ndarray)
    assert isinstance(data['Extent'], list)
    assert isinstance(data['Resolution'], list)
    assert isinstance(data['Nodata_val'], float)
    assert isinstance(data['Dimensions'], tuple)
    assert isinstance(data['CRS'], str)



"""
Contributors: Alexander Jüstel, Arthur Endlein Correia, Florian Wellmann

GemGIS is a Python-based, open-source geographic information processing library.
It is capable of preprocessing spatial data such as vector data (shape files, geojson files, geopackages),
raster data, data obtained from WMS services or XML/KML files.
Preprocessed data can be stored in a dedicated Data Class to be passed to the geomodeling package GemPy
in order to accelerate to model building process.

GemGIS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

GemGIS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (LICENSE.md) for more details.

"""

import io
import os
import numpy as np
from typing import Union, List
import geopandas as gpd


# Trying to import owslib but returning error if owslib is not installed
try:
    import owslib
    from owslib import util
    from owslib.wms import WebMapService
    from owslib.wfs import WebFeatureService
    from owslib.wcs import WebCoverageService
    __all__ = [util]
except ModuleNotFoundError:
    pass
    #raise ModuleNotFoundError('owslib package is not installed. Use pip install owslib to install the latest version')

# Trying to import requests but returning error if requests is not installed
try:
    import requests
except ModuleNotFoundError:
    raise ModuleNotFoundError('requests package is not installed. Use pip install requests to install the latest version')

# Working with Online Services
##############################

# Working with Web Mas Services
###############################


def load_wms(url: str) -> owslib.wms.WebMapService:
    """Loading a WMS Service by URL

    Parameters
    __________

         url : str
            Link of the WMS Service, e.g. ``url='https://ows.terrestris.de/osm/service?'``

    Returns
    _______

        wms : owslib.map.wms111.WebMapService
            OWSLib WebMapService Object

    Example
    _______

        >>> # Loading Libraries and WMS Service
        >>> import gemgis as gg
        >>> wms = gg.web.load_wms(url='https://ows.terrestris.de/osm/service?')
        >>> wms
        <owslib.map.wms111.WebMapService_1_1_1 at 0x1c434eb6370>


    See Also
    ________

        load_as_map : Load Map from WMS Service
        load_as_array : Load Map as array from WMS Service

    """

    # Checking if url is of type string
    if not isinstance(url, str):
        raise TypeError('URL must be of type string')

    # Requesting the WMS Service or returning an error if a module may be missing
    try:
        wms = owslib.wms.WebMapService(url)

        return wms
    except requests.exceptions.SSLError:
        print("GemGIS: SSL Error, potentially related to missing module - try:\n\n pip install -U openssl \n\n")
        raise


def load_as_map(url: str,
                layer: str,
                style: str,
                crs: Union[str, dict],
                bbox: List[Union[float, int]],
                size: List[int],
                filetype: str,
                transparent: bool = True,
                save_image: bool = False,
                path: str = None,
                overwrite_file: bool = False,
                create_directory: bool = False) -> owslib.util.ResponseWrapper:
    """Loading a portion of a WMS as array

    Parameters
    __________

        url : str
            Link of the WMS Service, e.g. ``url='https://ows.terrestris.de/osm/service?'``

        layer : str
            Name of layer to be requested, e.g. ``layer='OSM-WMS'``

        style : str
            Name of style of the layer, e.g. ``style='default'``

        crs : str
            String or dict containing the CRS, e.g. ``crs='EPSG:4647'``

        bbox : List[Union[float,int]]
            List of bounding box coordinates, e.g. ``bbox=[0, 972, 0, 1069]``

        size : List[int]
            List of x and y values defining the size of the image, e.g. ``size=[1000,1000]``

        filetype : str
           String of the image type to be downloaded, e.g. 'filetype='image/png'``

        transparent : bool
            Variable if layer is transparent.
            Options include: ``True`` or ``False``, default set to ``True``

        save_image : bool
            Variable to save image.
            Options include: ``True`` or ``False``, default set to ``False``

        path : str
            Path and file name of the file to be saved, e.g. ``path=map.tif``

        overwrite_file : bool
            Variable to overwrite an already existing file.
            Options include: ``True`` or ``False``, default set to ``False``

        create_directory : bool
            Variable to create a new directory of directory does not exist
            Options include: ``True`` or ``False``, default set to ``False``


    Returns
    _______

        wms_map : owslib.util.ResponseWrapper
             OWSlib map object

    Example
    _______

        >>> # Loading Libraries and WMS Service as Map
        >>> import gemgis as gg
        >>> wms_map = gg.web.load_as_map(url='https://ows.terrestris.de/osm/service?', layer='OSM-WMS', style='default', crs='EPSG:4647', bbox=[32286000,32328000, 5620000,5648000], size=[4200, 2800], filetype='image/png')
        >>> wms_map
        <owslib.util.ResponseWrapper at 0x261d348cc10>

    See Also
    ________

        load_wms : Load WMS Service
        load_as_array : Load Map as array from WMS Service

    """

    # Checking if the url is of type string
    if not isinstance(url, str):
        raise TypeError('URL must be of type string')

    # Checking if the layer name is of type string
    if not isinstance(layer, str):
        raise TypeError('Layers must be of type string')

    # Checking if the style is of type string
    if not isinstance(style, str):
        raise TypeError('Style must be of type string')

    # Checking if the crs is of type string or dict
    if not isinstance(crs, (str, dict)):
        raise TypeError('CRS must be of type str or dict')

    # Checking if bbox is of type list
    if not isinstance(bbox, list):
        raise TypeError('Bbox must be of type list')

    # Checking the length of the bbox list
    if len(bbox) != 4:
        raise ValueError('Provide minx, maxx, miny, and maxy values for the bounding box')

    # Checking if size is of type list
    if not isinstance(size, list):
        raise TypeError('Size must be of type list')

    # Checking the length of the size list
    if len(size) != 2:
        raise ValueError('Provide only a x- and y-value for the size')

    # Checking if file type is of type string
    if not isinstance(filetype, str):
        raise TypeError('File type must be of type string')

    # Checking if the transparency is of type book
    if not isinstance(transparent, bool):
        raise TypeError('transparent must be of type bool')

    # Checking if save_image is of type bool
    if not isinstance(save_image, bool):
        raise TypeError('Save_image must be of type bool')

    # Checking is path is of type string
    if not isinstance(path, (str, type(None))):
        raise TypeError('Path must be of type string')

    if isinstance(path, str):
        # Getting the absolute path
        path = os.path.abspath(path=path)

        # Checking that the file has the correct file ending
        if not path.endswith(".png"):
            if not path.endswith(".tif"):
                raise TypeError("The raster must be saved as .png or .tif file")

        # Getting path to directory
        path_dir = os.path.dirname(path)

        # Creating new directory
        if not os.path.exists(path_dir):
            if create_directory:
                os.makedirs(path_dir)
            else:
                raise LookupError('Directory not found. Pass create_directory=True to create a new directory')

        if not overwrite_file:
            if os.path.exists(path):
                raise FileExistsError(
                    "The file already exists. Pass overwrite_file=True to overwrite the existing file")

    # Loading WMS Service
    wms = load_wms(url)

    # Creating map object
    wms_map = wms.getmap(layers=[layer], styles=[style], srs=crs, bbox=tuple([bbox[0], bbox[2], bbox[1], bbox[3]]),
                         size=tuple(size), format=filetype,
                         transparent=transparent)

    # Saving an image if save_image is true and a path is provided
    if save_image:
        if isinstance(path, str):
            out = open(path, 'wb')
            out.write(wms_map.read())
            out.close()
        else:
            raise ValueError('Path is missing')
    else:
        if isinstance(path, str):
            raise ValueError('Save_image was set to False')

    return wms_map


def load_as_array(url: str,
                  layer: str,
                  style: str,
                  crs: Union[str, dict],
                  bbox: List[Union[float, int]],
                  size: List[int],
                  filetype: str,
                  transparent: bool = True,
                  save_image: bool = False,
                  path: str = None,
                  overwrite_file: bool = False,
                  create_directory: bool = False) -> np.ndarray:
    """Loading a portion of a WMS as array

    Parameters
    __________

        url : str
            Link of the WMS Service, e.g. ``url='https://ows.terrestris.de/osm/service?'``

        layer : str
            Name of layer to be requested, e.g. ``layer='OSM-WMS'``

        style : str
            Name of style of the layer, e.g. ``style='default'``

        crs : str
            String or dict containing the CRS, e.g. ``crs='EPSG:4647'``

        bbox : List[Union[float,int]]
            List of bounding box coordinates, e.g. ``bbox=[0, 972, 0, 1069]``

        size : List[int]
            List of x and y values defining the size of the image, e.g. ``size=[1000,1000]``

        filetype : str
           String of the image type to be downloaded, e.g. 'filetype='image/png'``

        transparent : bool
            Variable if layer is transparent.
            Options include: ``True`` or ``False``, default set to ``True``

        save_image : bool
            Variable to save image.
            Options include: ``True`` or ``False``, default set to ``False``

        path : str
            Path and file name of the file to be saved, e.g. ``path=map.tif``

        overwrite_file : bool
            Variable to overwrite an already existing file.
            Options include: ``True`` or ``False``, default set to ``False``

        create_directory : bool
            Variable to create a new directory of directory does not exist
            Options include: ``True`` or ``False``, default set to ``False``

    Returns
    _______

        wms_array: np.ndarray
             OWSlib map object loaded as np.ndarray

    Example
    _______

        >>> # Loading Libraries and WMS Service as array
        >>> import gemgis as gg
        >>> wms_map = gg.web.load_as_array(url='https://ows.terrestris.de/osm/service?', layer='OSM-WMS', style='default', crs='EPSG:4647', bbox=[32286000,32328000, 5620000,5648000], size=[4200, 2800], filetype='image/png')
        >>> wms_map
        array([[[0.8039216 , 0.7647059 , 0.65882355],
        [0.85882354, 0.8784314 , 0.6627451 ],
        [0.87058824, 0.91764706, 0.6666667 ],
        ...,
        [0.78431374, 0.7647059 , 0.65882355],
        [0.8862745 , 0.9019608 , 0.81960785],
        [0.9529412 , 0.93333334, 0.9019608 ]]], dtype=float32)

    See Also
    ________

        load_wms : Load WMS Service
        load_as_map : Load Map from WMS Service

    """

    # Trying to import matplotlib but returning error if matplotlib is not installed
    try:
        import matplotlib.pyplot as plt
    except ModuleNotFoundError:
        raise ModuleNotFoundError('Matplotlib package is not installed. Use pip install matplotlib to install the latest version')

    # Checking if the url is of type string
    if not isinstance(url, str):
        raise TypeError('URL must be of type string')

    # Checking if the layer name is of type string
    if not isinstance(layer, str):
        raise TypeError('Layers must be of type string')

    # Checking if the style is of type string
    if not isinstance(style, str):
        raise TypeError('Style must be of type string')

    # Checking if the crs is of type string or dict
    if not isinstance(crs, (str, dict)):
        raise TypeError('CRS must be of type str or dict')

    # Checking if bbox is of type list
    if not isinstance(bbox, list):
        raise TypeError('Bbox must be of type list')

    # Checking the length of the bbox list
    if len(bbox) != 4:
        raise ValueError('Provide minx, maxx, miny and maxy values for the bounding box')

    # Checking if size is of type list
    if not isinstance(size, list):
        raise TypeError('Size must be of type list')

    # Checking the length of the size list
    if len(size) != 2:
        raise ValueError('Provide only a x- and y-value for the size')

    # Checking if file type is of type string
    if not isinstance(filetype, str):
        raise TypeError('File type must be of type string')

    # Checking if the transparency is of type book
    if not isinstance(transparent, bool):
        raise TypeError('transparent must be of type bool')

    # Checking if save_image is of type bool
    if not isinstance(save_image, bool):
        raise TypeError('Save_image must be of type bool')

    # Checking is path is of type string
    if not isinstance(path, (str, type(None))):
        raise TypeError('Path must be of type string')

    if isinstance(path, str):
        # Getting the absolute path
        path = os.path.abspath(path=path)

        # Checking that the file has the correct file ending
        if not path.endswith(".png"):
            if not path.endswith(".tif"):
                raise TypeError("The raster must be saved as .png or .tif file")

        # Getting path to directory
        path_dir = os.path.dirname(path)

        # Creating new directory
        if not os.path.exists(path_dir):
            if create_directory:
                os.makedirs(path_dir)
            else:
                raise LookupError('Directory not found. Pass create_directory=True to create a new directory')

        if not overwrite_file:
            if os.path.exists(path):
                raise FileExistsError(
                    "The file already exists. Pass overwrite_file=True to overwrite the existing file")

    # Creating WMS map object
    wms_map = load_as_map(url=url,
                          layer=layer,
                          style=style,
                          crs=crs,
                          bbox=bbox,
                          size=size,
                          filetype=filetype,
                          transparent=transparent,
                          save_image=save_image,
                          path=path)

    # Converting WMS map object to array
    maps = io.BytesIO(wms_map.read())
    wms_array = plt.imread(maps)

    return wms_array


# Working with Web Feature Services
###################################


def load_wfs(url: str) -> owslib.wfs.WebFeatureService:
    """Loading an WMS Service by URL

    Parameters
    __________

         url : str
            Link of the WFS Service, e.g. ``url="https://nibis.lbeg.de/net3/public/ogc.ashx?NodeId=476&Service=WFS&"``

    Returns
    _______

        wfs : owslib.feature.wfs100.WebFeatureService_1_0_0
            OWSLib Feature object

    Example
    _______

        >>> # Loading Libraries and WFS Service
        >>> import gemgis as gg
        >>> wfs = gg.web.load_wfs(url="https://nibis.lbeg.de/net3/public/ogc.ashx?NodeId=476&Service=WFS&")
        >>> wfs
        <owslib.feature.wfs100.WebFeatureService_1_0_0 at 0x19260e21340>

    See Also
    ________

        load_as_gpd : Load information of a WFS Service as GeoDataFrame

    """

    # Checking if url is of type string
    if not isinstance(url, str):
        raise TypeError('URL must be of type string')

    # Requesting the WMS Service or returning an error if a module may be missing
    try:
        wfs = owslib.wfs.WebFeatureService(url)

        return wfs

    except requests.exceptions.SSLError:
        print("GemGIS: SSL Error, potentially related to missing module - try:\n\n pip install -U openssl \n\n")
        raise


def load_as_gpd(url: str,
                typename: str = None,
                outputformat: str = None
                ) -> gpd.geodataframe.GeoDataFrame:
    """Requesting data from a WFS Service

    Parameters
    __________

        url : str
            Url of the Web Feature Service, e.g. ``url="https://nibis.lbeg.de/net3/public/ogc.ashx?NodeId=476&Service=WFS&"``

        typename : str
            Name of the feature layer, e.g. ``typename='iwan:L383'``

        outputformat : str
            Output format of the feature layer, e.g. ``outputformat='xml/gml2'``

    Returns
    _______

        feature : gpd.geodataframe.GeoDataFrame
            GeoDataFrame containing the feature data of the WFS Service

    Example
    _______

        >>> # Loading Libraries and WFS Service as GeoDataFrame
        >>> import gemgis as gg
        >>> wfs = gg.web.load_as_gpd(url="https://nibis.lbeg.de/net3/public/ogc.ashx?NodeId=476&Service=WFS&")
        >>> wfs
            gml_id	OBJECTID    ID	SURVEYNAME	ARCHIV	MESSJAHR	OPERATOR	                                OP_NACHFOL	                    MESSFIRMA	                            MESSPUNKTE	UP_DATE	                    geometry
        0	1541	1541	    112	Jemgum 2007	0127494	2007	    GdF Produktion Exploration Deutschland GmbH	Neptune Energy Deutschland GmbH	Geophysik und Geotechnik Leipzig GmbH	1340	    2020-01-20T00:00:00+01:00	MULTIPOLYGON (((32395246.839 5907777.660, 3239...

    See Also
    ________

        load_wfs : Load WFS Service

    """

    # Checking that the url is of type string
    if not isinstance(url, str):
        raise TypeError('URL must be of type string')

    # Checking that the typename is of type string or None
    if not isinstance(typename, (str, type(None))):
        raise TypeError('Name of the feature must be of type string')

    # Checking that the outputformat is of type string
    if not isinstance(outputformat, (str, type(None))):
        raise TypeError('The output format must be of type string')

    # Loading the wfs layer
    wfs = load_wfs(url=url)

    # If the layer name is not provided, take the last layer of the service
    if not typename:
        layer = list(wfs.contents)[0]
    else:
        raise ValueError('No layer available')

    # If the output format is not provided, take the last
    if not outputformat:
        if wfs.getOperationByName('GetFeature').formatOptions == ['{http://www.opengis.net/wfs}GML2']:
            outputformat = 'xml/gml2'

    # Specify the parameters for fetching the data
    params = dict(service='WFS', version=wfs.version, request='GetFeature',
                  typeName=layer, outputFormat=outputformat)

    # Parse the URL with parameters
    q = requests.Request('GET', url, params=params).prepare().url

    # Read data from request
    feature = gpd.read_file(q)

    return feature


# Working with Web Coverage Services
####################################


def load_wcs(url: str) -> owslib.wcs.WebCoverageService:
    """Loading Web Coverage Service

    Parameters
    __________

        url : str
            Link of the Web Coverage Service, e.g. ``url='https://www.wcs.nrw.de/geobasis/wcs_nw_dgm'``

    Returns
    _______

        wcs : owslib.coverage.wcs201.WebCoverageService_2_0_1
            OWSLib Web Coverage Object

    Example
    _______

        >>> # Loading Libraries and WCS Service
        >>> import gemgis as gg
        >>> wcs = gg.web.load_wms(url='https://www.wcs.nrw.de/geobasis/wcs_nw_dgm')
        >>> wcs
        <owslib.coverage.wcs201.WebCoverageService_2_0_1 at 0x27fc64783d0>

    See Also
    ________

        create_request : Create request for WCS
        load_as_file : Download WCS data file
        load_as_files : Download WCS data files

    """

    # Checking if URL is of type string
    if not isinstance(url, str):
        raise TypeError('URL must be of type string')

    # Loading the WCS Layer
    wcs = owslib.wcs.WebCoverageService(url)

    return wcs


def create_request(wcs_url: str,
                   version: str,
                   identifier: str,
                   form: str,
                   extent: List[Union[float, int]],
                   name: str = 'test.tif') -> str:
    """Create URL to request data from WCS Server

    Parameters
    __________

        wcs_url : str
            Url of the WCS server, e.g. ``url='https://www.wcs.nrw.de/geobasis/wcs_nw_dgm'``

        version : str
            Version number of the WCS as string, e.g. ``version='2.0.1'``

        identifier : str
            Name of the layer, e.g. ``identifier='nw_dgm'``

        form : str
            Format of the layer, e.g. ``form='image/tiff'``

        extent : List[Union[float,int]]
            Extent of the tile to be downloaded, size may be restricted by server,
            e.g. ``extent=[0, 972, 0, 1069]``

        name : str
            Name of file, e.g. ``name='tile1'``

    Returns
    _______

        url : str
            Url for the WCS request

    Example
    _______

        >>> # Loading Libraries and WCS Service
        >>> import gemgis as gg
        >>> wcs = gg.web.load_wms(url='https://www.wcs.nrw.de/geobasis/wcs_nw_dgm')
        >>> wcs
        <owslib.coverage.wcs201.WebCoverageService_2_0_1 at 0x27fc64783d0>

        >>> # Creating Request for WCS Service
        >>> url = gg.web.create_request(url=wcs.url, version=wcs.version, identifier='nw_dgm', form='image/tiff', extent=[0, 1000, 0, 1000], name='test.tif'])

    See Also
    ________

        load_wcs : Load WCS Service
        load_as_file : Download WCS data file
        load_as_files : Download WCS data files

    """

    # Checking that the URL is of type string
    if not isinstance(wcs_url, str):
        raise TypeError('URL must be of type string')

    # Checking that the version number is of type string
    if not isinstance(version, str):
        raise TypeError('WCS Version must be of type string')

    # Checking that the identifier is of type string
    if not isinstance(identifier, str):
        raise TypeError('Layer Name/Identifier must be of type string')

    # Checking that the format is of type string
    if not isinstance(form, str):
        raise TypeError('Download format must be of type string')

    # Checking that the extent is of type list
    if not isinstance(extent, list):
        raise TypeError('Extent must be provided as list of minx, maxx, miny, maxy')

    # Checking the length of the extent
    if len(extent) != 4:
        raise ValueError('Extent must be provided as list of minx, maxx, miny, maxy')

    # Create URL for Request
    url = wcs_url + '?' + 'REQUEST=GetCoverage' + '&' + 'SERVICE=WCS' + '&' + 'VERSION=' + str(version) + '&' + \
                          'COVERAGEID=' + identifier + '&' + 'FORMAT=' + form + '&' + \
                          'SUBSET=x(' + str(extent[0]) + ',' + str(extent[1]) + ')' + '&' + \
                          'SUBSET=y(' + str(extent[2]) + ',' + str(extent[3]) + ')' + '&' + 'OUTFILE=' + name

    return url


def load_as_file(url: str,
                 path: str,
                 overwrite_file: bool = False,
                 create_directory: bool = False):
    """Executing WCS request and downloading file into specified folder

    Parameters
    __________

        url: str
            Url for request

        path: str
            Path where file is saved, e.g. ``path='tile.tif'``

        overwrite_file : bool
            Variable to overwrite an already existing file.
            Options include: ``True`` or ``False``, default set to ``False``

        create_directory : bool
            Variable to create a new directory of directory does not exist
            Options include: ``True`` or ``False``, default set to ``False``

    Example
    _______

        >>> # Loading Libraries and WCS Service
        >>> import gemgis as gg
        >>> wcs = gg.web.load_wms(url='https://www.wcs.nrw.de/geobasis/wcs_nw_dgm')
        >>> wcs
        <owslib.coverage.wcs201.WebCoverageService_2_0_1 at 0x27fc64783d0>

        >>> # Creating Request for WCS Service
        >>> url = gg.web.create_request(url=wcs.url, version=wcs.version, identifier='nw_dgm', form='image/tiff', extent=[0, 1000, 0, 1000], name='test.tif'])

        >>> # Downloading file from WCS Service
        >>> gg.web.load_as_file(url=url, path='tile.tif')

    See Also
    ________

        load_wcs : Load WCS Service
        create_request : Create request for WCS
        load_as_files : Download WCS data files

    """

    # Trying to import urllib but returning error if urllib is not installed
    try:
        import urllib
    except ModuleNotFoundError:
        raise ModuleNotFoundError('urllib package is not installed. Use pip install urllib to install the latest version')

    # Checking that the url is of type string
    if not isinstance(url, str):
        raise TypeError('URL must be of type string')

    # Checking that the path is of type string
    if not isinstance(path, str):
        raise TypeError('Path must be of type string')

    # Getting the absolute path
    path = os.path.abspath(path=path)

    # Checking that the file has the correct file ending
    if not path.endswith(".tif"):
        raise TypeError("The raster must be saved as .tif file")

    # Getting path to directory
    path_dir = os.path.dirname(path)

    # Creating new directory
    if not os.path.exists(path_dir):
        if create_directory:
            os.makedirs(path_dir)
        else:
            raise LookupError('Directory not found. Pass create_directory=True to create a new directory')

    if not overwrite_file:
        if os.path.exists(path):
            raise FileExistsError(
                "The file already exists. Pass overwrite_file=True to overwrite the existing file")

    # Executing request and downloading files to the specified folder
    urllib.request.urlretrieve(url, path)


def load_as_files(wcs_url: str,
                  version: str,
                  identifier: str,
                  form: str,
                  extent: List[Union[float, int]],
                  size: int,
                  path: str = '',
                  create_directory: bool = False):
    """Executing WCS requests and downloading files into specified folder

    Parameters
    __________

        wcs_url : str
            Url of the WCS server, e.g. ``url='https://www.wcs.nrw.de/geobasis/wcs_nw_dgm'``

        version : str
            Version number of the WCS as string, e.g. ``version='2.0.1'``

        identifier : str
            Name of the layer, e.g. ``identifier='nw_dgm'``

        form : str
            Format of the layer, e.g. ``form='image/tiff'``

        extent : List[Union[float,int]]
            Extent of the tile to be downloaded, size may be restricted by server,
            e.g. ``extent=[0, 972, 0, 1069]``

        size : int
            Size of the quadratic tile that is downloaded, e.g. ``size=2000``

        path : str
            Path where the file is going to be downloaded, e.g. ``name='tile1'``

        create_directory : bool
            Variable to create a new directory of directory does not exist
            Options include: ``True`` or ``False``, default set to ``False``

    Example
    _______

        >>> # Loading Libraries and WCS Service
        >>> import gemgis as gg
        >>> wcs = gg.web.load_wms(url='https://www.wcs.nrw.de/geobasis/wcs_nw_dgm')
        >>> wcs
        <owslib.coverage.wcs201.WebCoverageService_2_0_1 at 0x27fc64783d0>

        >>> # Downloading files from WCS Service
        >>> gg.web.load_as_files(wcs_url=wcs.url, version=wcs.version, form='image/tiff', extent=[0, 10000, 0, 10000], size=2000, path='tile.tif')

    See Also
    ________

        load_wcs : Load WCS Service
        create_request : Create request for WCS
        load_as_file : Download WCS data file

    """

    # Trying to import tqdm but returning error if tqdm is not installed
    try:
        from tqdm import tqdm
    except ModuleNotFoundError:
        raise ModuleNotFoundError('tqdm package is not installed. Use pip install tqdm to install the latest version')

    # Checking that the URL is of type string
    if not isinstance(wcs_url, str):
        raise TypeError('URL must be of type string')

    # Checking that the version number is of type string
    if not isinstance(version, str):
        raise TypeError('WCS Version must be of type string')

    # Checking that the identifier is of type string
    if not isinstance(identifier, str):
        raise TypeError('Layer Name/Identifier must be of type string')

    # Checking that the format is of type string
    if not isinstance(form, str):
        raise TypeError('Download format must be of type string')

    # Checking that the extent is of type list
    if not isinstance(extent, list):
        raise TypeError('Extent must be provided as list of minx, maxx, miny, maxy')

    # Checking the length of the extent
    if len(extent) != 4:
        raise ValueError('Extent must be provided as list of minx, maxx, miny, maxy')

    # Checking that the provided size of each tile is of type int
    if not isinstance(size, int):
        raise TypeError('Tile size must be provided as int')

    # Calculating the x Extent
    x = extent[1] - extent[0]

    # Calculating the y extent
    y = extent[3] - extent[2]

    # Printing the extent and number of tiles that are going to be downloaded
    print('Extent X: ', x, ' m')
    print('Extent Y: ', y, ' m')
    print('Number of tiles in X directions: ', int(x / size))
    print('Number of tiles in Y directions: ', int(y / size))
    print('Total Number of Tiles: ', int(x / size) * int(y / size))

    # Loop through each tile and download data
    for i in tqdm(range(int(x / size))):
        for j in range(int(y / size)):
            # Download data only if the tile does not exist yet
            if not os.path.exists(path + 'tile_%d_%d_%d_%d.tif' %
                                  (extent[0] + i * size,
                                   extent[0] + (i + 1) * size,
                                   extent[2] + j * size,
                                   extent[2] + (j + 1) * size)):

                # Create URL request
                url = create_request(wcs_url=wcs_url,
                                     version=version,
                                     identifier=identifier,
                                     form=form,
                                     extent=[extent[0] + i * size,
                                             extent[0] + (i + 1) * size,
                                             extent[2] + j * size,
                                             extent[2] + (j + 1) * size],
                                     name=path)
                print(url)

                # Load file
                load_as_file(url=url,
                             path=path + 'tile_%d_%d_%d_%d.tif' % (extent[0] + i * size,
                                                                   extent[0] + (i + 1) * size,
                                                                   extent[2] + j * size,
                                                                   extent[2] + (j + 1) * size),
                             create_directory=create_directory)
            else:
                print('All tiles have already been downloaded')
                pass



"""
Contributors: Alexander Jüstel, Arthur Endlein Correia, Florian Wellmann

GemGIS is a Python-based, open-source geographic information processing library.
It is capable of preprocessing spatial data such as vector data (shape files, geojson files,
geopackages), raster data (tif, png,...), data obtained from web services (WMS, WFS, WCS) or XML/KML
files. Preprocessed data can be stored in a dedicated Data Class to be passed to the geomodeling package
GemPy in order to accelerate to model building process. In addition, enhanced 3D visualization of data is
powered by the PyVista package.

GemGIS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

GemGIS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (LICENSE.md) for more details.

"""

import pytest
import geopandas as gpd
import rasterio
import pyvista as pv
import numpy as np
import pandas as pd
from shapely.geometry import LineString, Point

points1 = Point(19.150128045807676, 293.313485355882)

points2 = Point(61.93436666575576, 381.4593263680641)

points3 = Point(109.3578600758187, 480.9455679783049)

points4 = Point(157.812298994796, 615.9994296460927)

points5 = Point(191.3180280345144, 719.0939805375339)

gdf_interfaces1_points = gpd.GeoDataFrame(geometry=[points1, points2, points3, points4, points5], crs='EPSG:4326')
gdf_interfaces1_points['formation'] = 'Ton'
gdf_interfaces1_points['id'] = None

lines1 = LineString([[0.7408806771479846, 475.4410147469845],
                     [35.62873136073459, 429.2469161566801],
                     [77.30033078835194, 340.0890755208477],
                     [104.7583614189525, 269.3442671902416],
                     [127.0478215779106, 207.6444571850097]])

lines2 = LineString([[645.9649974657704, 0.524960824189406],
                     [685.1409268045179, 61.8662186045969],
                     [724.8323288977228, 114.4444395592319],
                     [753.6988031473263, 158.7750964425515],
                     [781.5343318880155, 194.858189254556]])

lines3 = LineString([[490.2922256196942, 0.5249608241894066],
                     [505.7564082534104, 40.73183567185146],
                     [519.1586998692977, 79.39229225614187],
                     [529.4681549584418, 95.88742039877246],
                     [575.8607028595903, 158.2596236880943]])

gdf_topo1 = gpd.GeoDataFrame(geometry=[lines1, lines2, lines3], crs='EPSG:4326')
gdf_topo1['Z'] = [400, 300, 400]
gdf_topo1['id'] = None

points1 = Point(96.47104121438838, 451.5636209742439)

points2 = Point(172.7610088740548, 661.8765047927839)

points3 = Point(383.0738926925949, 957.7578658512201)

points4 = Point(592.3558310022205, 722.7022898187342)

points5 = Point(766.5856220087561, 348.4690700828027)

points6 = Point(843.906535177337, 167.0226605138662)

points7 = Point(941.8463585242062, 428.8828197781268)

points8 = Point(22.14220797837953, 299.5527565162123)

gdf_orientations1 = gpd.GeoDataFrame(
    geometry=[points1, points2, points3, points4, points5, points6, points7, points8, ], crs='EPSG:4326')
gdf_orientations1['id'] = None
gdf_orientations1['formation'] = 'Ton'
gdf_orientations1['dip'] = 30.5
gdf_orientations1['azimuth'] = 180
gdf_orientations1['X'] = [96.47104121438838, 172.76100887405482, 383.0738926925949, 592.3558310022205,
                          766.5856220087561, 843.906535177337, 941.8463585242062, 22.14220797837953]
gdf_orientations1['Y'] = [451.5636209742439, 661.8765047927839, 957.7578658512201, 722.7022898187342,
                          348.46907008280266, 167.02266051386619, 428.88281977812676, 299.5527565162123]
gdf_orientations1['Z'] = [477.72849453315575,481.7329984389544, 444.4466863869966, 480.5679521803904,
                          498.95859477613703, 537.8202132122024, 511.95775336451516, 446.6623707773251]


# Testing plot_points_3d
###########################################################
@pytest.mark.parametrize("gdf_interfaces1_points", [gdf_interfaces1_points])
@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_plot_points_3d(gdf_interfaces1_points, dem):
    from gemgis.visualization import create_points_3d
    from gemgis.vector import extract_xyz

    gdf = extract_xyz(gdf=gdf_interfaces1_points,
                      dem=dem)

    assert isinstance(gdf, gpd.geodataframe.GeoDataFrame)
    assert {'X', 'Y', 'Z'}.issubset(gdf.columns)

    points = create_points_3d(gdf=gdf)

    assert isinstance(points, pv.core.pointset.PolyData)


@pytest.mark.parametrize("gdf_interfaces1_points", [gdf_interfaces1_points])
@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_plot_points_3d_error(gdf_interfaces1_points, dem):
    from gemgis.visualization import create_points_3d
    from gemgis.vector import extract_xyz

    gdf = extract_xyz(gdf=gdf_interfaces1_points,
                      dem=dem)

    assert isinstance(gdf, gpd.geodataframe.GeoDataFrame)
    assert {'X', 'Y', 'Z'}.issubset(gdf.columns)

    with pytest.raises(TypeError):
        create_points_3d(gdf=[gdf])
    with pytest.raises(TypeError):
        create_points_3d(gdf=np.array(gdf))
    with pytest.raises(TypeError):
        create_points_3d(gdf='gdf')


# Testing plot_dem_3d
###########################################################
@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_plot_dem_3d(dem):
    from gemgis.visualization import create_dem_3d

    mesh = create_dem_3d(dem=dem,
                         extent=[0, 250, 0, 275])

    assert isinstance(mesh, pv.core.pointset.StructuredGrid)


@pytest.mark.parametrize("dem",
                         [
                             rasterio.open('../../gemgis/tests/data/raster1.tif')
                         ])
def test_plot_dem_3d_error(dem):
    from gemgis.visualization import create_dem_3d

    create_dem_3d(dem=dem,
                  extent=[0, 250, 0, 275])

    with pytest.raises(TypeError):
        create_dem_3d(dem=[dem], extent=[0, 250, 0, 275])
    with pytest.raises(TypeError):
        create_dem_3d(dem=dem, extent=(0, 250, 0, 275))
    with pytest.raises(TypeError):
        create_dem_3d(dem='dem', extent=[0, 250, 0, 275])


# Testing plot_contours_3d
###########################################################
@pytest.mark.parametrize("gdf_topo1", [gdf_topo1])
def test_plot_contours_3d(gdf_topo1):
    from gemgis.visualization import create_lines_3d

    mesh = create_lines_3d(gdf=gdf_topo1)

    assert isinstance(mesh, pv.core.pointset.PolyData)


@pytest.mark.parametrize("gdf_topo1", [gdf_topo1])
def test_plot_contours_3d_error(gdf_topo1):
    from gemgis.visualization import create_lines_3d

    with pytest.raises(TypeError):
        create_lines_3d(gdf=[gdf_topo1])
    with pytest.raises(TypeError):
        create_lines_3d(gdf='gdf_topo1')
    with pytest.raises(TypeError):
        create_lines_3d(gdf=np.array(gdf_topo1))


# Testing create_mesh_cross_section
###########################################################
def test_create_mesh_cross_section():
    from gemgis.visualization import create_mesh_from_cross_section

    trace = LineString([(0, 0), (10, 10)])

    mesh = create_mesh_from_cross_section(linestring=trace,
                                          zmax=0,
                                          zmin=-10)

    assert isinstance(mesh, pv.core.pointset.PolyData)


# Testing create_meshes_from_cross_sections
###########################################################
def test_create_meshes_from_cross_sections():
    from gemgis.visualization import create_meshes_from_cross_sections

    trace = LineString([[0, 0], [10, 10]])

    gdf = gpd.GeoDataFrame(geometry=[trace, trace])
    gdf['zmax'] = 0
    gdf['zmin'] = -10

    meshes = create_meshes_from_cross_sections(gdf=gdf)

    assert isinstance(meshes, list)
    assert all(isinstance(n, pv.core.pointset.PolyData) for n in meshes)


# Testing read_raster
###########################################################
def test_read_raster():
    from gemgis.visualization import read_raster

    mesh = read_raster(path='../../gemgis/tests/data/raster1.tif',
                       nodata_val=None,
                       name='Elevation [m]')

    assert isinstance(mesh, pv.core.pointset.StructuredGrid)


# Testing convert_to_rgb
###########################################################
@pytest.mark.parametrize("array",
                         [
                             np.load('../../gemgis/tests/data/wms.npy')
                         ])
def test_convert_to_rgb(array):
    from gemgis.visualization import convert_to_rgb

    array_rgb = convert_to_rgb(array)

    assert isinstance(array_rgb, np.ndarray)


# Testing drape_array_over_dem
###########################################################
# @pytest.mark.parametrize("array",
#                          [
#                              np.load('../../gemgis/tests/data/wms.npy')
#                          ])
# @pytest.mark.parametrize("dem",
#                          [
#                              rasterio.open('../../gemgis/tests/data/DEM50.tif')
#                          ])
# def test_drape_array_over_dem(array, dem):
#     from gemgis.visualization import convert_to_rgb, drape_array_over_dem
#
#     array_rgb = convert_to_rgb(array)
#
#     assert isinstance(array_rgb, np.ndarray)
#
#     mesh, texture = drape_array_over_dem(array=array,
#                                          dem=dem)
#
#     assert isinstance(mesh, pv.core.pointset.StructuredGrid)
#     assert isinstance(texture, pv.core.objects.Texture)


# Testing create_polydata_from_msh
###########################################################
def test_create_polydata_from_msh():
    from gemgis.raster import read_msh
    from gemgis.visualization import create_polydata_from_msh

    data = read_msh('../../gemgis/tests/data/GM_Breccia.msh')

    mesh = create_polydata_from_msh(data=data)

    assert isinstance(mesh, pv.core.pointset.PolyData)


# Testing create_polydata_from_ts
###########################################################
def test_create_polydata_from_ts():
    from gemgis.raster import read_ts
    from gemgis.visualization import create_polydata_from_ts

    data = read_ts('../../gemgis/tests/data/KVB_12_Hermann_Katharina.ts')

    mesh = create_polydata_from_ts(data=data)

    assert isinstance(mesh, pv.core.pointset.PolyData)


# Testing create_depth_maps
###########################################################
# @pytest.mark.parametrize("gdf_interfaces1_points", [gdf_interfaces1_points])
# @pytest.mark.parametrize("gdf_orientations1", [gdf_orientations1])
# def test_create_depth_maps_from_gempy(gdf_interfaces1_points, gdf_orientations1):
#     from gemgis.visualization import create_depth_maps_from_gempy
#     import gempy as gp
#     import pandas as pd
#
#     extent = [0, 972, 0, 1069, 300, 800]
#     resolution = [50, 50, 50]
#     gdf_orientations1['polarity'] = 1
#     geo_model = gp.create_model('Model1')
#
#     gp.init_data(geo_model, extent, resolution,
#                  surface_points_df=pd.DataFrame(gdf_interfaces1_points.drop('geometry', axis=1)),
#                  orientations_df=pd.DataFrame(gdf_orientations1.drop('geometry', axis=1)),
#                  default_values=True)
#
#     gp.map_stack_to_surfaces(geo_model,
#                              {"Strat_Series": ('Sand1', 'Ton')},
#                              remove_unused_series=True)
#     geo_model.add_surfaces('basement')
#
#     geo_model.set_topography(
#         source='gdal', filepath='../../gemgis/tests/data/raster1.tif')
#
#     gp.set_interpolator(geo_model,
#                         compile_theano=True,
#                         theano_optimizer='fast_compile',
#                         verbose=[],
#                         update_kriging=False
#                         )
#
#     gp.compute_model(geo_model, compute_mesh=True)
#
#     dict_all = create_depth_maps_from_gempy(geo_model=geo_model,
#                                             surfaces=['Sand1', 'Ton'])
#
#     assert isinstance(dict_all, dict)
#     assert len(dict_all) == 2
#     assert 'Sand1' in dict_all
#     assert 'Ton' in dict_all
#     assert isinstance(dict_all['Sand1'][0], pv.core.pointset.PolyData)
#     assert isinstance(dict_all['Sand1'][1], str)
#
#
# # Testing create_thickness maps
# ###########################################################
# @pytest.mark.parametrize("gdf_interfaces1_points", [gdf_interfaces1_points])
# @pytest.mark.parametrize("gdf_orientations1", [gdf_orientations1])
# def test_create_thickness_maps(gdf_interfaces1_points, gdf_orientations1):
#     from gemgis.visualization import create_depth_maps_from_gempy, create_thickness_maps
#     import gempy as gp
#     import pandas as pd
#
#     extent = [0, 972, 0, 1069, 300, 800]
#     resolution = [50, 50, 50]
#     gdf_orientations1['polarity'] = 1
#     geo_model = gp.create_model('Model1')
#
#     gp.init_data(geo_model, extent, resolution,
#                  surface_points_df=pd.DataFrame(gdf_interfaces1_points.drop('geometry', axis=1)),
#                  orientations_df=pd.DataFrame(gdf_orientations1.drop('geometry', axis=1)),
#                  default_values=True)
#
#     gp.map_stack_to_surfaces(geo_model,
#                              {"Strat_Series": ('Sand1', 'Ton')},
#                              remove_unused_series=True)
#     geo_model.add_surfaces('basement')
#
#     geo_model.set_topography(
#         source='gdal', filepath='../../gemgis/tests/data/raster1.tif')
#
#     gp.set_interpolator(geo_model,
#                         compile_theano=True,
#                         theano_optimizer='fast_compile',
#                         verbose=[],
#                         update_kriging=False
#                         )
#
#     gp.compute_model(geo_model, compute_mesh=True)
#
#     dict_all = create_depth_maps_from_gempy(geo_model=geo_model,
#                                             surfaces=['Sand1', 'Ton'])
#
#     assert isinstance(dict_all, dict)
#     assert len(dict_all) == 2
#     assert 'Sand1' in dict_all
#     assert 'Ton' in dict_all
#     assert isinstance(dict_all['Sand1'][0], pv.core.pointset.PolyData)
#     assert isinstance(dict_all['Sand1'][1], str)
#
#     thickness_map = create_thickness_maps(top_surface=dict_all['Sand1'][0],
#                                           base_surface=dict_all['Ton'][0])
#
#     assert isinstance(thickness_map, pv.core.pointset.PolyData)
#

# Testing create_depth_map
###########################################################
@pytest.mark.parametrize('mesh',
                         [
                             pv.read('../../gemgis/tests/data/mesh1.vtk')
                         ])
def test_create_depth_map(mesh):
    from gemgis.visualization import create_depth_map

    mesh = create_depth_map(mesh=mesh)

    assert isinstance(mesh, pv.core.pointset.PolyData)
    assert 'Depth [m]' in mesh.array_names


# Testing create_meshes_hypocenters
###########################################################
@pytest.mark.parametrize('data',
                         [
                             gpd.read_file('../../gemgis/tests/data/earthquake_data.shp')
                         ])
def test_create_meshes_hypocenters(data):
    from gemgis.visualization import create_meshes_hypocenters

    spheres = create_meshes_hypocenters(gdf=data)

    assert isinstance(spheres, pv.core.composite.MultiBlock)


# Testing create_polydata_from_dxf
###########################################################
# @pytest.mark.parametrize('gdf',
#                          [
#                              gpd.read_file('../../gemgis/tests/data/Channel.dxf')
#                          ])
# def test_create_polydata_from_dxf(gdf):
#     from gemgis.visualization import create_polydata_from_dxf
#
#     polydata = create_polydata_from_dxf(gdf=gdf)
#
#     assert isinstance(polydata, pv.core.pointset.PolyData)


# Testing plane_through_hypocenters
###########################################################
@pytest.mark.parametrize('spheres',
                         [
                             pv.read('../../gemgis/tests/data/spheres.vtm')
                         ])
def test_plane_through_hypocenters(spheres):
    from gemgis.visualization import plane_through_hypocenters

    plane = plane_through_hypocenters(spheres=spheres)

    assert isinstance(plane, pv.core.pointset.PolyData)


# # Testing create_structured_grid_from_asc
# ###########################################################
# def test_create_structured_grid_from_asc():
#     from gemgis.raster import read_asc
#     from gemgis.visualization import create_structured_grid_from_asc
#
#     data = read_asc(path='../../gemgis/tests/data/top_dinant_final_tvd.asc')
#
#     assert isinstance(data['Data'], np.ndarray)
#     assert isinstance(data['Extent'], list)
#     assert isinstance(data['Resolution'], float)
#     assert isinstance(data['Nodata_val'], float)
#
#     grid = create_structured_grid_from_asc(data=data)
#
#     assert isinstance(grid, pv.core.pointset.StructuredGrid)


# # Testing create_structured_grid_from_zmap
# ###########################################################
# def test_create_structured_grid_from_zmap():
#     from gemgis.raster import read_zmap
#     from gemgis.visualization import create_structured_grid_from_zmap
#
#     data = read_zmap(path='../../gemgis/tests/data/top_dinant_final_tvd.dat')
#
#     assert isinstance(data['Data'], np.ndarray)
#     assert isinstance(data['Extent'], list)
#     assert isinstance(data['Resolution'], list)
#     assert isinstance(data['Nodata_val'], float)
#     assert isinstance(data['Dimensions'], tuple)
#     assert isinstance(data['CRS'], str)
#
#     grid = create_structured_grid_from_zmap(data=data)
#
#     assert isinstance(grid, pv.core.pointset.StructuredGrid)


# Testing calculate_vector
###########################################################
def test_calculate_vector():
    from gemgis.visualization import calculate_vector

    vector = calculate_vector(dip=90, azimuth=20)

    assert isinstance(vector, np.ndarray)


# Testing create_deviated_borehole_df
###########################################################
@pytest.mark.parametrize('collar',
                         [
                             pd.read_csv('../../gemgis/tests/data/collar.csv', delimiter=';')
                         ])
@pytest.mark.parametrize('survey',
                         [
                             pd.read_csv('../../gemgis/tests/data/survey.csv')
                         ])
def test_create_deviated_borehole_df(collar, survey):
    from gemgis.visualization import create_deviated_borehole_df

    survey006 = survey[survey['holeid'] == 'SonicS_006']
    survey006 = survey006.reset_index().drop('index', axis=1)

    x0 = collar[['x', 'y', 'z']].loc[2].values

    df_survey = create_deviated_borehole_df(df_survey=survey006, position=x0)

    assert isinstance(df_survey, pd.DataFrame)


# Testing create_deviated_boreholes_3d
###########################################################
@pytest.mark.parametrize('collar',
                         [
                             pd.read_csv('../../gemgis/tests/data/collar.csv', delimiter=';')
                         ])
@pytest.mark.parametrize('survey',
                         [
                             pd.read_csv('../../gemgis/tests/data/survey.csv')
                         ])
def test_create_deviated_boreholes_3d(collar, survey):
    from gemgis.visualization import create_deviated_boreholes_3d

    tubes, df_groups = create_deviated_boreholes_3d(df_collar=collar,
                                                    df_survey=survey,
                                                    min_length=10,
                                                    collar_depth='maxdepth',
                                                    survey_depth='depth',
                                                    index='holeid')

    assert isinstance(tubes, pv.core.composite.MultiBlock)
    assert isinstance(df_groups, list)


"""
Contributors: Alexander Jüstel, Arthur Endlein Correia, Florian Wellmann

GemGIS is a Python-based, open-source geographic information processing library.
It is capable of preprocessing spatial data such as vector data (shape files, geojson files,
geopackages), raster data (tif, png,...), data obtained from web services (WMS, WFS, WCS) or XML/KML
files. Preprocessed data can be stored in a dedicated Data Class to be passed to the geomodeling package
GemPy in order to accelerate to model building process. In addition, enhanced 3D visualization of data is
powered by the PyVista package.

GemGIS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

GemGIS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (LICENSE.md) for more details.

"""

import pandas as pd
import geopandas as gpd
import re


# Testing load_pdf
###########################################################
def test_load_pdf():
    from gemgis.misc import load_pdf

    pdf = load_pdf(path='../../gemgis_data/data/tests/test_pdf.pdf',
                   save_as_txt=True)

    assert isinstance(pdf, str)


# Testing get_coordinate_data
###########################################################
def test_get_meta_data():
    from gemgis.misc import get_meta_data, load_pdf

    pdf = load_pdf(path='../../gemgis_data/data/tests/test_pdf.pdf',
                   save_as_txt=False)

    data = pdf.split()
    data = '#'.join(data)
    data = data.split('-Stammdaten')
    data = [item.split('|')[0] for item in data]
    data = [item.split('#') for item in data]

    # Filter out wells without Stratigraphic Column
    data = [item for item in data if 'Beschreibung' in item]

    data = [get_meta_data(page=item) for item in data]

    assert isinstance(data, list)


# Testing coordinates_table_list_comprehension
###########################################################
def test_get_meta_data_df():
    from gemgis.misc import get_meta_data_df, load_pdf

    pdf = load_pdf(path='../../gemgis_data/data/tests/test_pdf.pdf')

    assert isinstance(pdf, str)

    df = get_meta_data_df(data=pdf,
                          name='Test',
                          return_gdf=False)

    assert isinstance(df, pd.DataFrame)
    assert len(df) == 2
    assert df.loc[0]['Depth'] == 1242
    assert df.loc[1]['Depth'] == 1135
    assert df.loc[0]['Name'] == 'ASCHEBERG12STK.'
    assert df.loc[1]['Name'] == 'ASCHEBERG15STK.'
    assert df.loc[0]['X'] == 32407673.17
    assert df.loc[1]['X'] == 32407713.16
    assert df.loc[0]['Y'] == 5742123.75
    assert df.loc[1]['Y'] == 5742143.75
    assert df.loc[0]['Z'] == 60
    assert df.loc[1]['Z'] == 60

    df = get_meta_data_df(data=pdf,
                          name='Test',
                          return_gdf=True)

    assert isinstance(df, gpd.geodataframe.GeoDataFrame)
    assert len(df) == 2
    assert df.loc[0]['Depth'] == 1242
    assert df.loc[1]['Depth'] == 1135
    assert df.loc[0]['Name'] == 'ASCHEBERG12STK.'
    assert df.loc[1]['Name'] == 'ASCHEBERG15STK.'
    assert df.loc[0]['X'] == 32407673.17
    assert df.loc[1]['X'] == 32407713.16
    assert df.loc[0]['Y'] == 5742123.75
    assert df.loc[1]['Y'] == 5742143.75
    assert df.loc[0]['Z'] == 60
    assert df.loc[1]['Z'] == 60


# Testing get_stratigraphic_data
###########################################################
def test_get_stratigraphic_data():
    from gemgis.misc import get_stratigraphic_data, load_pdf

    pdf = load_pdf(path='../../gemgis_data/data/tests/test_pdf.pdf')

    assert isinstance(pdf, str)

    with open('../../gemgis_data/data/tests/symbols.txt', "r") as text_file:
        symbols = [(i, '') for i in text_file.read().splitlines()]

    with open('../../gemgis_data/data/tests/formations.txt', "rb") as text_file:
        formations = text_file.read().decode("UTF-8").split()

    formations = [(formations[i], formations[i + 1]) for i in range(0, len(formations) - 1, 2)]

    # Splitting the entire string into a list
    data = pdf.split()

    # Join all elements of list/all pages of the borehole logs and separate with #
    data = '#'.join(data)

    # Split entire string at each new page into separate elements of a list
    data = data.split('-Stammdaten')

    # Cut off the last part of each element, this is not done for each page
    data = [item.split('|Geologischer#Dienst#NRW#')[0] for item in data]

    # Remove last part of each page if log stretches over multiple pages
    data = [re.sub('Geologischer#Dienst#NRW#\d\d.\d\d.\d\d\d\d-#\d+#-#', '#', item) for item in data]
    data = [re.sub('Geologischer#Dienst#NRW#\d\d.\d\d.\d\d\d\d-#\d+#-', '#', item) for item in data]

    # Connect different parts of each element
    data = [''.join(item) for item in data]

    # Split each element at #
    data = [item.split('#') for item in data]

    # Filter out wells without Stratigraphic Column
    data = [item for item in data if 'Beschreibung' in item]

    # Get stratigraphic data for each well
    stratigraphy = [get_stratigraphic_data(text=item,
                                           symbols=symbols,
                                           formations=formations) for item in data]

    assert isinstance(stratigraphy, list)


# Testing get_stratigraphic_data_df
###########################################################
def test_get_stratigraphic_data_df():
    from gemgis.misc import get_stratigraphic_data_df, load_pdf

    pdf = load_pdf(path='../../gemgis_data/data/tests/test_pdf.pdf')

    assert isinstance(pdf, str)

    with open('../../gemgis_data/data/tests/symbols.txt', "r") as text_file:
        symbols = [(i, '') for i in text_file.read().splitlines()]

    with open('../../gemgis_data/data/tests/formations.txt', "rb") as text_file:
        formations = text_file.read().decode("UTF-8").split()

    formations = [(formations[i], formations[i + 1]) for i in range(0, len(formations) - 1, 2)]

    df = get_stratigraphic_data_df(data=pdf,
                                   name='GD',
                                   symbols=symbols,
                                   formations=formations,
                                   return_gdf=False)

    assert isinstance(df, pd.DataFrame)

    df = get_stratigraphic_data_df(data=pdf,
                                   name='GD',
                                   symbols=symbols,
                                   formations=formations,
                                   return_gdf=True)

    assert isinstance(df, gpd.geodataframe.GeoDataFrame)


# Testing stratigraphic_table_list_comprehension
###########################################################
def test_stratigraphic_table_list_comprehension():
    from gemgis.misc import get_stratigraphic_data_df, load_pdf

    with open('../../gemgis_data/data/tests/symbols.txt', "r") as text_file:
        symbols = [(i, '') for i in text_file.read().splitlines()]

    with open('../../gemgis_data/data/tests/formations.txt', "rb") as text_file:
        formations = text_file.read().decode("UTF-8").split()

    formations = [(formations[i], formations[i + 1]) for i in range(0, len(formations) - 1, 2)]

    pdf = load_pdf('../../gemgis_data/data/tests/test_pdf.pdf')

    assert type(pdf) == str

    df = get_stratigraphic_data_df(data=pdf,
                                   name='Test',
                                   symbols=symbols,
                                   formations=formations,
                                   return_gdf=False)

    assert type(df) == pd.DataFrame
    assert len(df) == 7
    assert df.loc[0]['Depth'] == 1242
    assert df.loc[4]['Depth'] == 1135
    assert df.loc[0]['Name'] == 'ASCHEBERG12STK.'
    assert df.loc[4]['Name'] == 'ASCHEBERG15STK.'
    assert df.loc[0]['X'] == 32407673.17
    assert df.loc[4]['X'] == 32407713.16
    assert df.loc[0]['Y'] == 5742123.75
    assert df.loc[4]['Y'] == 5742143.75
    assert df.loc[0]['Z'] == -870
    assert df.loc[4]['Z'] == 59.5
    assert df.loc[0]['Altitude'] == 60
    assert df.loc[4]['Altitude'] == 60


# Testing load_symbols
###########################################################
def test_load_symbols():
    from gemgis.misc import load_symbols

    symbols = load_symbols(path='../../gemgis_data/data/tests/symbols20201216.txt')

    assert isinstance(symbols, list)


# Testing load_formations
###########################################################
def test_load_formations():
    from gemgis.misc import load_formations

    formations = load_formations(path='../../gemgis_data/data/tests/formations20210109.txt')

    assert isinstance(formations, list)



"""
Contributors: Alexander Jüstel, Arthur Endlein Correia, Florian Wellmann

GemGIS is a Python-based, open-source geographic information processing library.
It is capable of preprocessing spatial data such as vector data (shape files, geojson files, geopackages),
raster data, data obtained from WMS services or XML/KML files.
Preprocessed data can be stored in a dedicated Data Class to be passed to the geomodeling package GemPy
in order to accelerate to model building process.

GemGIS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

GemGIS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (LICENSE.md) for more details.

"""

import numpy as np
# import scooby
import pandas as pd
from pandas.core import frame
import rasterio
import geopandas as gpd
import rasterio.transform
from typing import Union
from gemgis.vector import extract_xy, extract_xyz
from gemgis.utils import parse_categorized_qml, build_style_dict
from gemgis.raster import calculate_hillshades, calculate_slope, calculate_aspect
from gemgis.utils import create_surface_color_dict


# class Report(scooby.Report):
#     def __init__(self, additional=None, ncol=3, text_width=80, sort=False):
#         """Initiate a scooby.Report instance."""
#
#         # Mandatory packages.
#         core = ['json', 'numpy', 'scooby', 'owslib', 'pandas', 'shapely', 'pyvista', 'rasterio', 'geopandas',
#                 'requests', 'scipy', 'skimage']
#
#         # Optional packages.
#         optional = ['your_optional_packages', 'e.g.', 'matplotlib']
#
#         scooby.Report.__init__(self, additional=additional, core=core,
#                                optional=optional, ncol=ncol,
#                                text_width=text_width, sort=sort)


# Class tested
class GemPyData(object):
    """
    This class creates an object with attributes containing i.e. the interfaces or orientations
    that can directly be passed to a GemPy Model
   
    The following attributes are available:
    - model_name: string - the name of the model
    - crs: string - the coordinate reference system of the model
    - interfaces: pd DataFrame - DataFrame containing the interfaces for the GemPy model
    - orientations: pd DataFrame - DataFrame containing the orientations for the GemPy model
    - extent: list - List containing the minx, maxx, miny, maxy, minz and maxz values
    - section_dict: dict - Dictionary containing the section_dict for custom sections for the GemPy model
    - customsections: GeoDataFrame containing the Linestrings or Endpoints of custom sections
    - resolution: list - List containing the x,y and z resolution of the model
    - dem: Union[string, array] - String containing the path to the DEM or array containing DEM values
    - stack: dict - Dictionary containing the layer stack associated with the model 
    - surface_colors: dict - Dictionary containing the surface colors for the model 
    - is_fault: list - list of surface that are classified as faults
    - geolmap: Union[GeoDataFrame,np.ndarray rasterio.io.Datasetreader] - GeoDataFrame or array containing
    the geological map either as vector or raster data set
    - basemap: Union[np.ndarray rasterio.io.Datasetreader] - Array or rasterio object containing a base map of the area
    - tectonics: GeoDataFrame - GeoDataFrame containing the LineStrings of fault traces
    - raw_i: GeoDataFrame - GeoDataFrame containing the raw interfaces point data
    - raw_o: GeoDataFrame - GeoDataFrame containing the raw orientations data
    - raw_dem: GeoDataFrame or np.ndarray - Raw dem data such as topographic lines or (gdf) or raster (array)
    - slope: np.ndarray - array containing the slope values of the DEM
    - hillshades: np.ndarray - array containing the color values of the hillshades
    - aspect: np.ndarray - array containing the aspect values of the DEM
    - faults: GeoDataFrame containing the Linestrings or vertices of faults
    - wms: np.ndarray containing data obtained from a WMS layer
    - contours: GeoDataFrame containing the contour lines of the model area
    """

    def __init__(self,
                 model_name=None,
                 crs=None,
                 extent=None,
                 resolution=None,
                 interfaces=None,
                 orientations=None,
                 section_dict=None,
                 customsections=None,
                 dem=None,
                 stack=None,
                 surface_colors=None,
                 is_fault=None,
                 geolmap=None,
                 basemap=None,
                 faults=None,
                 tectonics=None,
                 raw_i=None,
                 raw_o=None,
                 raw_dem=None,
                 wms=None,
                 slope=None,
                 hillshades=None,
                 aspect=None,
                 contours=None):

        # Checking if data type are correct

        # Checking if model name was provided as string
        if isinstance(model_name, (type(None), str)):
            self.model_name = model_name
        else:
            raise TypeError("Model Name must be of type str")

        # Checking if CRS was provided as string
        if isinstance(crs, (type(None), str)):
            self.crs = crs
        else:
            raise TypeError("CRS must be of type str")

        # Checking if extent was provided as list of 6 elements (int/floats)
        if isinstance(extent, (type(None), list)):
            if isinstance(extent, list):
                if len(extent) == 6:
                    if all(isinstance(n, (int, (int, float))) for n in extent):
                        self.extent = extent
                    else:
                        raise TypeError('Coordinates for extent must be provided as integers or floats')
                else:
                    raise ValueError('Length of extent must be 6 [minx,maxx,miny,maxy,minz,maxz]')
            self.extent = extent
        else:
            raise TypeError("Extent must be of type list")

        # Checking if the resolution was provided as list of 3 integers
        if isinstance(resolution, (type(None), list)):
            if isinstance(resolution, list):
                if len(resolution) == 3:
                    if all(isinstance(n, int) for n in resolution):
                        self.resolution = resolution
                    else:
                        raise TypeError('Values for resolution must be provided as integers')
                else:
                    raise ValueError('Length of resolution must be 3 [x,y,z]')
            self.resolution = resolution
        else:
            raise TypeError("Resolution must be of type list")

        # Checking if the interfaces object is a Pandas df containing all relevant columns
        if isinstance(interfaces, (type(None), pd.core.frame.DataFrame)):
            if isinstance(interfaces, pd.core.frame.DataFrame):
                assert pd.Series(['X', 'Y', 'Z', 'formation']).isin(
                    interfaces.columns).all(), 'Interfaces DataFrame is missing columns'
            self.interfaces = interfaces
        else:
            raise TypeError("Interfaces df must be a Pandas DataFrame")

        # Checking if the orientations object is Pandas df containing all relevant columns
        if isinstance(orientations, (type(None), pd.core.frame.DataFrame)):
            if isinstance(orientations, pd.core.frame.DataFrame):
                assert pd.Series(['X', 'Y', 'Z', 'formation', 'dip', 'azimuth', 'polarity']).isin(
                    orientations.columns).all(), 'Orientations DataFrame is missing columns'
            self.orientations = orientations
        else:
            raise TypeError("Orientations df must be a Pandas DataFrame")

        # Setting the section_dict attribute
        if isinstance(section_dict, (type(None), dict)):
            self.section_dict = section_dict
        else:
            raise TypeError("Section Dict must be of type dict")

        # Checking if the provided stack is of type dict
        if isinstance(stack, (type(None), dict)):
            self.stack = stack
        else:
            raise TypeError("Layer Stack must be of type dict")

        # Checking if the provided DEM object is either an np array, a file loaded with rasterio or a string
        if isinstance(dem, (type(None), np.ndarray, rasterio.io.DatasetReader, str)):
            self.dem = dem
        else:
            raise TypeError("Digital Elevation Model must be a np Array, a raster loaded with rasterio or a string")

        # Checking if the provided surface colors object is of type dict
        if isinstance(surface_colors, (type(None), dict)):
            self.surface_colors = surface_colors
        elif isinstance(surface_colors, str):
            self.surface_colors = create_surface_color_dict('../../gemgis/data/Test1/style1.qml')
        else:
            raise TypeError("Surface Colors Dict must be of type dict or a path directing to a qml file")

        # Checking that the provided geological map is a gdf containing polygons
        if isinstance(geolmap, (type(None), gpd.geodataframe.GeoDataFrame, rasterio.io.DatasetReader, np.ndarray)):
            if isinstance(geolmap, gpd.geodataframe.GeoDataFrame):
                if all(geolmap.geom_type == "Polygon"):
                    self.geolmap = geolmap
                else:
                    raise TypeError("Geometry Type must be Polygon")
            elif isinstance(geolmap, rasterio.io.DatasetReader):
                self.geolmap = geolmap.read(1)
            else:
                self.geolmap = geolmap
        else:
            raise TypeError("Geological Map must be a GeoDataFrame or NumPy Array")

        # Checking that the provided basemap is a np.ndarray or rasterio data set
        if isinstance(basemap, (type(None), rasterio.io.DatasetReader, np.ndarray)):
            if isinstance(basemap, rasterio.io.DatasetReader):
                self.basemap = basemap.read(1)
            else:
                self.basemap = basemap
        else:
            raise TypeError('Base Map must be a Raster loaded with rasterio or a NumPy Array')

        # Checking the the provided faults are a gdf containing LineStrings
        if isinstance(faults, (type(None), gpd.geodataframe.GeoDataFrame)):
            if isinstance(faults, gpd.geodataframe.GeoDataFrame):
                if all(faults.geom_type == "LineString"):
                    self.faults = faults
                else:
                    raise TypeError("Geometry Type must be LineString")
            self.faults = faults
        else:
            raise TypeError("Faults must be a GeoDataFrame")

        # Checking that the provided is_fault object is a list containing strings
        if isinstance(is_fault, (type(None), list)):
            if isinstance(is_fault, list):
                if all(isinstance(n, str) for n in is_fault):
                    self.is_fault = is_fault
                else:
                    raise TypeError('Fault Names must be provided as strings')
            self.is_fault = is_fault
        else:
            TypeError('List of faults must be of type list')

        # Checking that the provided raw input data objects are of type gdf
        if isinstance(raw_i, (gpd.geodataframe.GeoDataFrame, type(None))):
            self.raw_i = raw_i
        if isinstance(raw_o, (gpd.geodataframe.GeoDataFrame, type(None))):
            self.raw_o = raw_o
        if isinstance(raw_dem, (gpd.geodataframe.GeoDataFrame, np.ndarray, type(None))):
            self.raw_dem = raw_dem

        # Setting the slope attribute
        if isinstance(slope, np.ndarray):
            self.slope = slope
        elif isinstance(self.raw_dem, np.ndarray) and isinstance(slope, type(None)):
            self.slope = calculate_slope(self.raw_dem, self.extent)
        else:
            self.slope = slope

        # Setting the hillshades attribute
        if isinstance(hillshades, np.ndarray):
            self.hillshades = hillshades
        elif isinstance(self.raw_dem, np.ndarray) and isinstance(hillshades, type(None)):
            self.hillshades = calculate_hillshades(self.raw_dem, self.extent)
        else:
            self.hillshades = hillshades

        # Setting the aspect attribute
        if isinstance(aspect, np.ndarray):
            self.aspect = aspect
        elif isinstance(self.raw_dem, np.ndarray) and isinstance(aspect, type(None)):
            self.aspect = calculate_aspect(self.raw_dem, self.extent)
        else:
            self.aspect = aspect

        # Calculate model dimensions
        if not isinstance(self.extent, type(None)):
            self.model_width = self.extent[1]-self.extent[0]
            self.model_length = self.extent[3]-self.extent[2]
            self.model_depth = self.extent[5]-self.extent[4]
            self.model_area = self.model_width*self.model_length
            self.model_volume = self.model_area*self.model_depth

        # Calculate cell dimensions
        if not isinstance(self.resolution, type(None)):
            if not isinstance(self.extent, type(None)):
                self.cell_width = self.model_width/self.resolution[0]
                self.cell_length = self.model_length/self.resolution[1]
                self.cell_depth = self.model_depth/self.resolution[2]

        # Setting the wms attribute
        if isinstance(wms, np.ndarray):
            self.wms = wms
        else:
            self.wms = None

        # Setting the tectonics attribute
        self.tectonics = tectonics

        # Setting the customsections attribute
        if isinstance(customsections, (type(None), gpd.geodataframe.GeoDataFrame)):
            if isinstance(customsections, gpd.geodataframe.GeoDataFrame):
                self.customsections = customsections
            else:
                self.customsections = customsections
        else:
            raise TypeError('Custom sections must be provided as GeoDataFrame')

        # Setting the contours attribute
        if isinstance(contours, (type(None), gpd.geodataframe.GeoDataFrame)):
            if isinstance(contours, gpd.geodataframe.GeoDataFrame):
                self.contours = contours
            else:
                self.contours = contours
        else:
            raise TypeError('Custom sections must be provided as GeoDataFrame')

    # Function tested
    def to_section_dict(self, gdf: gpd.geodataframe.GeoDataFrame, section_column: str = 'section_name',
                        resolution=None):
        """
        Converting custom sections stored in shape files to GemPy section_dicts
        Args:
            gdf - gpd.geodataframe.GeoDataFrame containing the points or lines of custom sections
            section_column - string containing the name of the column containing the section names
            resolution - list containing the x,y resolution of the custom section
        Return:
             section_dict containing the section names, coordinates and resolution
        """

        if resolution is None:
            resolution = [100, 80]

        # Checking if gdf is of type GeoDataFrame
        if not isinstance(gdf, gpd.geodataframe.GeoDataFrame):
            raise TypeError('gdf must be of type GeoDataFrame')

        # Checking if the section_column is of type string
        if not isinstance(section_column, str):
            raise TypeError('Name for section_column must be of type string')

        # Checking if resolution is of type list
        if not isinstance(resolution, list):
            raise TypeError('resolution must be of type list')

        # Checking if X and Y values are in column
        if np.logical_not(pd.Series(['X', 'Y']).isin(gdf.columns).all()):
            gdf = extract_xy(gdf)

        # Checking the length of the resolution list
        if len(resolution) != 2:
            raise ValueError('resolution list must be of length two')

        # Checking that a valid section name is provided
        if not {'section_name'}.issubset(gdf.columns):
            if not {section_column}.issubset(gdf.columns):
                raise ValueError('Section_column name needed to create section_dict')

        # Extracting Section names
        section_names = gdf[section_column].unique()

        # Create section dicts for Point Shape Files
        if all(gdf.geom_type == "Point"):
            section_dict = {i: ([gdf[gdf[section_column] == i].X.iloc[0], gdf[gdf[section_column] == i].Y.iloc[0]],
                                [gdf[gdf[section_column] == i].X.iloc[1], gdf[gdf[section_column] == i].Y.iloc[1]],
                                resolution) for i in section_names}

        # Create section dicts for Line Shape Files
        else:
            section_dict = {i: ([gdf[gdf[section_column] == i].X.iloc[0], gdf[gdf[section_column] == i].Y.iloc[0]],
                                [gdf[gdf[section_column] == i].X.iloc[1], gdf[gdf[section_column] == i].Y.iloc[1]],
                                resolution) for i in section_names}

        self.section_dict = section_dict

    # Function tested
    def to_gempy_df(self, gdf: gpd.geodataframe.GeoDataFrame, cat: str, **kwargs):
        """
        Converting a GeoDataFrame into a Pandas DataFrame ready to be read in for GemPy
        Args:
            gdf - gpd.geodataframe.GeoDataFrame containing spatial information, formation names and orientation values
            cat - str/type of point data (interfaces or orientations)
        Kwargs:
            dem -
        Return:
             df - interface or orientations DataFrame ready to be read in for GemPy
        """

        # Checking if gdf is of type GeoDataFrame
        if not isinstance(gdf, gpd.geodataframe.GeoDataFrame):
            raise TypeError('gdf must be of type GeoDataFrame')

        # Checking if type is of type string
        if not isinstance(cat, str):
            raise TypeError('Type must be of type string')

        if np.logical_not(pd.Series(['X', 'Y', 'Z']).isin(gdf.columns).all()):
            dem = kwargs.get('dem', None)
            extent = kwargs.get('extent', None)

            if not isinstance(dem, type(None)):
                gdf = extract_xyz(gdf, dem, extent=extent)
            else:
                raise FileNotFoundError('DEM not provided to obtain Z values for point data')
        if np.logical_not(pd.Series(['formation']).isin(gdf.columns).all()):
            raise ValueError('formation names not defined')

        # Converting dip and azimuth columns to floats
        if pd.Series(['dip']).isin(gdf.columns).all():
            gdf['dip'] = gdf['dip'].astype(float)

        if pd.Series(['azimuth']).isin(gdf.columns).all():
            gdf['azimuth'] = gdf['azimuth'].astype(float)

        # Converting formation column to string
        if pd.Series(['formation']).isin(gdf.columns).all():
            gdf['formation'] = gdf['formation'].astype(str)

        # Checking if dataframe is an orientation or interfaces df
        if pd.Series(['dip']).isin(gdf.columns).all():
            if cat == 'orientations':
                if (gdf['dip'] > 90).any():
                    raise ValueError('dip values exceed 90 degrees')
                if np.logical_not(pd.Series(['azimuth']).isin(gdf.columns).all()):
                    raise ValueError('azimuth values not defined')
                if (gdf['azimuth'] > 360).any():
                    raise ValueError('azimuth values exceed 360 degrees')

                # Create orientations dataframe
                if np.logical_not(pd.Series(['polarity']).isin(gdf.columns).all()):
                    df = pd.DataFrame(gdf[['X', 'Y', 'Z', 'formation', 'dip', 'azimuth']].reset_index())
                    df['polarity'] = 1
                    self.orientations = df
                else:
                    self.orientations = \
                        pd.DataFrame(gdf[['X', 'Y', 'Z', 'formation', 'dip', 'azimuth', 'polarity']].reset_index())
            else:
                raise ValueError('GeoDataFrame contains orientations but type is interfaces')
        else:
            if cat == 'interfaces':
                # Create interfaces dataframe
                self.interfaces = pd.DataFrame(gdf[['X', 'Y', 'Z', 'formation']].reset_index())
            else:
                raise ValueError('GeoDataFrame contains interfaces but type is orientations')

    # Function tested
    def set_extent(self, minx: Union[int, float] = 0,
                   maxx: Union[int, float] = 0,
                   miny: Union[int, float] = 0,
                   maxy: Union[int, float] = 0,
                   minz: Union[int, float] = 0,
                   maxz: Union[int, float] = 0,
                   **kwargs):
        """
            Setting the extent for a model
            Args:
                minx - float defining the left border of the model
                maxx - float defining the right border of the model
                miny - float defining the upper border of the model
                maxy - float defining the lower border of the model
                minz - float defining the top border of the model
                maxz - float defining the bottom border of the model
            Kwargs:
                gdf - GeoDataFrame from which bounds the extent will be set
            Return:
                extent - list with resolution values
            """
        gdf = kwargs.get('gdf', None)

        if not isinstance(gdf, (type(None), gpd.geodataframe.GeoDataFrame)):
            raise TypeError('gdf mus be of type GeoDataFrame')

        # Checking if bounds are of type int or float
        if not all(isinstance(i, (int, float)) for i in [minx, maxx, miny, maxy, minz, maxz]):
            raise TypeError('bounds must be of type int or float')

        # Checking if the gdf is of type None
        if isinstance(gdf, type(None)):
            if minz == 0 and maxz == 0:
                extent = [minx, maxx, miny, maxy]
            else:
                extent = [minx, maxx, miny, maxy, minz, maxz]
        # Create extent from gdf of geom_type polygon
        elif all(gdf.geom_type == "Polygon"):
            # Checking if the gdf is of type GeoDataFrame
            bounds = gdf.bounds.round().values.tolist()[0]
            extent = [bounds[0], bounds[2], bounds[1], bounds[3], minz, maxz]
        # Create extent from gdf of geom_type point or linestring
        else:
            bounds = gdf.bounds
            extent = [round(bounds.minx.min(), 2), round(bounds.maxx.max(), 2), round(bounds.miny.min(), 2),
                      round(bounds.maxy.max(), 2), minz, maxz]

        self.extent = extent
        self.model_width = self.extent[1] - self.extent[0]
        self.model_length = self.extent[3] - self.extent[2]
        if len(self.extent) == 6:
            self.model_depth = self.extent[5] - self.extent[4]
        else:
            self.model_depth = 0
        self.model_area = self.model_width * self.model_length
        self.model_volume = self.model_area * self.model_depth

    # Function tested
    def set_resolution(self, x: int, y: int, z: int):
        """
        Setting the resolution for a model
        Args:
            x - int defining the resolution in X direction
            y - int defining the resolution in Y direction
            z - int defining the resolution in Z direction
        Return:
            [x, y, z] - list with resolution values
        """

        # Checking if x is of type int
        if not isinstance(x, int):
            raise TypeError('X must be of type int')

        # Checking if y is of type int
        if not isinstance(y, int):
            raise TypeError('Y must be of type int')

        # Checking if y is of type int
        if not isinstance(z, int):
            raise TypeError('Z must be of type int')

        self.resolution = [x, y, z]

        if not isinstance(self.extent, type(None)):
            self.cell_width = self.model_width / self.resolution[0]
            self.cell_length = self.model_length / self.resolution[1]
            self.cell_depth = self.model_depth / self.resolution[2]

    # Function tested
    def to_surface_color_dict(self, path: str, **kwargs):
        """
        Create GemPy surface color dict from a qml file
        Args:
            path: str/path to the qml file
        Return:
            surface_color_dict: dict containing the surface color values for GemPy
        """

        # Checking that the path is of type str
        if not isinstance(path, str):
            raise TypeError('path must be provided as string')

        # Parse qml
        columns, classes = parse_categorized_qml(path)

        # Create Styles
        styles = build_style_dict(classes)

        # Create surface_colors_dict
        surface_colors_dict = {k: v["color"] for k, v in styles.items() if k}

        basement = kwargs.get('basement', None)

        # Checking if discarded formation is of type string
        if not isinstance(basement, (str, type(None))):
            raise TypeError('Basement formation name must be of type string')

        # Pop oldest lithology from dict as it does not need a color in GemPy
        if isinstance(basement, str):
            surface_colors_dict['basement'] = surface_colors_dict.pop(basement)

        self.surface_colors = surface_colors_dict
